# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)
import os
import numpy as np
import json

import grid2op
from chronix2grid.grid2op_utils.utils import generate_a_scenario
from numpy.random import default_rng


def generate_one_episode(env: grid2op.Environment.Environment,
                         dict_ref,
                         dt,
                         start_date,
                         nb_steps=None,
                         seed=None,
                         with_loss=True):
    """This function adds some data to already existing scenarios.
    
    .. warning::
        You should not start this function twice. Before starting a new run, make sure the previous one has terminated (otherwise you might
        erase some previously generated scenario)

    Parameters
    ----------
    env : _type_
        The grid2op environment
    seed:
        The seed to use (the same seed is guaranteed to generate the same scenarios)
    nb_scenario: ``int``
        The number of scenarios to generate
    nb_core: ``int``
        The number of core you want to use (to speed up the generation process)
    with_loss: ``bool``
        Do you make sure that the generated data will not be modified too much when running with grid2op (default = True).
        Setting it to False will speed up (by quite a lot) the generation process, but will degrade the data quality.
        
    """
    # generate the seeds
    if seed is not None:
        prng = default_rng(seed)
    else:
        prng = default_rng()
    
    load_seed, renew_seed, gen_p_forecast_seed = prng.integers(2**32 - 1, size=3)
    path_env = env.get_path_env()
    name_gen = env.name_gen
    gen_type = env.gen_type
    scen_id = "0"
    error_ = "first"
    output_dir = None
    while error_ is not None:
        res_gen = generate_a_scenario(path_env, name_gen, gen_type, output_dir, start_date, dt, scen_id, load_seed, renew_seed, 
                                      gen_p_forecast_seed, with_loss, nb_steps=nb_steps)
        error_, quality_, load_p, load_p_forecasted, load_q, load_q_forecasted, res_gen_p_df, res_gen_p_forecasted_df = res_gen
    return load_p, load_p_forecasted, load_q, load_q_forecasted, res_gen_p_df, res_gen_p_forecasted_df



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import argparse
import numpy as np
import pandas as pd
import seaborn as sns
from itertools import product
from pathlib import Path
from matplotlib import pyplot as plt


class WindKPI:

    def __init__(self, kpi_databanks):
        self.kpi_databanks = kpi_databanks

    def _pairwise_corr_different_dfs(self, df1, df2):

        n_col_df1 = df1.shape[1]
        n_col_df2 = df2.shape[1]

        tmp_corr = np.zeros((n_col_df1, n_col_df2))
        for i, j in product(range(n_col_df1), range(n_col_df2)):
            tmp_corr[i, j] = df1.iloc[:, i].corr(df2.iloc[:, j])

        corr_wind = pd.DataFrame(tmp_corr, index=df1.columns, columns=df2.columns)

        return corr_wind.round(self.precision)

    def __wind_metric_distrib(self, wind_df):

        '''
        Return:
        -------
        Skewness: measures simmetry
        Kurtosis: measures tailedness
        '''

        # Add month as column to compute stats
        copied_wind_df = wind_df.copy()
        # wind_df_month['month'] = self.months

        # Get Skewness agg per month
        skewness_per_month = copied_wind_df.groupby(pd.Grouper(freq='M')).apply(lambda x: x.skew()).round(2)
        # Set index as month value
        skewness_per_month.index = skewness_per_month.index.month
        skewness_per_month.index.rename('month', inplace=True)

        # Get Kurtosis agg per month
        kurtosis_per_month = copied_wind_df.groupby(pd.Grouper(freq='M')).apply(lambda x: x.kurtosis()).round(2)
        # Set index as month value
        kurtosis_per_month.index = kurtosis_per_month.index.month
        kurtosis_per_month.index.rename('month', inplace=True)

        return skewness_per_month, kurtosis_per_month

    def wind_kpi(self, save_plots=False):

        '''
        Return:
        -------

        fig (plot): Subplot (1x2) containing the aggregated wind production

        corr_wind (pd.dataFrame): Correlation matrix (10x10) between ref data
                                  and synthethic one
        '''

        # First KPI
        # Correlation between wind time series

        # Get the wind gen names
        wind_filter = self.prod_charac.type.isin(['wind'])
        wind_names = self.prod_charac.name[wind_filter]

        # From dispatch, get only wind units
        wind_ref = self.ref_dispatch[wind_names]
        wind_syn = self.syn_dispatch[wind_names]

        # Compute correlation for all elements between both dataframes
        corr_wind = self._pairwise_corr_different_dfs(wind_ref, wind_syn)

        # Write results json output
        # -- + -- + -- + -- + -- +
        self.output['wind_kpi'] = {'corr_wind': corr_wind.to_dict()}

        # Second KPI
        # Measure non linearity of time series
        chaoticness_ref = self.__wind_entropy(wind_ref)
        chaoticness_syn = self.__wind_entropy(wind_syn)

        # Write results
        # -- + -- + --
        self.output['wind_kpi'] = {'non_linearity_reference': chaoticness_ref.to_dict(),
                                   'non_linearity_synthetic': chaoticness_syn.to_dict(),
                                   }

        # Third KPI
        # Meaure the simmetry of wind distribution
        skewness_ref, kurtosis_ref = self.__wind_metric_distrib(wind_ref)
        skewness_syn, kurtosis_syn = self.__wind_metric_distrib(wind_syn)

        # Write results
        # -- + -- + --
        self.output['wind_kpi'] = {'skewness_reference': skewness_ref.to_dict(),
                                   'skewness_synthetic': skewness_syn.to_dict(),
                                   'kurtosis_reference': kurtosis_ref.to_dict(),
                                   'kurtosis_synthetic': kurtosis_syn.to_dict(),
                                   }

        # Four KPI
        # Plot distributions

        # Aggregate time series
        agg_ref_wind = wind_ref.sum(axis=1)
        agg_syn_wind = wind_syn.sum(axis=1)

        # Plot results
        fig, axes = plt.subplots(1, 3, figsize=(17, 5))
        sns.heatmap(corr_wind, annot=True, linewidths=.5, ax=axes[0])
        sns.distplot(agg_ref_wind, ax=axes[1])
        sns.distplot(agg_syn_wind, ax=axes[2])
        axes[1].set_title('Reference Distribution')
        axes[2].set_title('Synthetic Distribution')

        if save_plots:
            # Save plot as pnd
            extent0 = axes[0].get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            extent1 = axes[1].get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            extent2 = axes[2].get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            fig.savefig('images/wind_kpi/win_corr_heatmap.png', bbox_inches=extent0.expanded(1.6, 1.6))
            fig.savefig('images/wind_kpi/ref_histogram.png', bbox_inches=extent1.expanded(1.4, 1.4))
            fig.savefig('images/wind_kpi/syn_histogram.png', bbox_inches=extent2.expanded(1.4, 1.4))

        return corr_wind, \
               chaoticness_ref, chaoticness_syn, \
               skewness_ref, skewness_syn, \
               kurtosis_ref, kurtosis_syn

    def wind_load_kpi(self):

        '''
        Return:
        ------

        corr(win, load) Region R1 (pd.DataFrame)
        corr(win, load) Region R2 (pd.DataFrame)
        corr(win, load) Region R3 (pd.DataFrame)
        '''

        # Get the solar gen names
        wind_filter = self.prod_charac.type.isin(['wind'])

        regions = ['R1', 'R2', 'R3']
        corr_rel = []
        for region in regions:
            # Create zone filter for wind and loads
            wind_region_filter = self.prod_charac.zone.isin([region])
            wind_names = self.prod_charac.loc[wind_filter & wind_region_filter]['name']
            loads_names_filter = self.load_charac[self.load_charac.zone == region]['name']

            # Extract only wind units per region
            wind_gens_in_region = self.syn_dispatch[wind_names]
            loads_in_region = self.consumption[loads_names_filter]

            # Compute correlation matrix
            tmp_corr = self._pairwise_corr_different_dfs(wind_gens_in_region, loads_in_region)
            corr_rel.append(tmp_corr)

        # Plot results
        plt.figure(figsize=(18, 4))
        self._plot_heatmap(corr_rel[0],
                           'Correlation Wind Load Region R1',
                           path_png='images/wind_load_kpi/corr_wind_load_R1.png',
                           save_png=False)

        plt.figure(figsize=(18, 4))
        self._plot_heatmap(corr_rel[1],
                           'Correlation Wind Load Region R2',
                           path_png='images/wind_load_kpi/corr_wind_load_R2.png',
                           save_png=False)

        plt.figure(figsize=(12, 3))
        self._plot_heatmap(corr_rel[2],
                           'Correlation Wind Load Region R3',
                           path_png='images/wind_load_kpi/corr_wind_load_R3.png',
                           save_png=False)

        return corr_rel[0], corr_rel[1], corr_rel[2]



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import argparse
import numpy as np
import pandas as pd
import seaborn as sns
from itertools import product
from pathlib import Path
from matplotlib import pyplot as plt

class NuclearKPI:

    def __init__(self, kpi_databanks):
        self.kpi_databanks = kpi_databanks

    def nuclear_kpi(self, save_plots=False):

        """
        Return:
        ------

        None
        """

        # Get the nuclear gen names
        nuclear_filter = self.prod_charac.type.isin(['nuclear'])
        nuclear_names = self.prod_charac.name.loc[nuclear_filter].values

        # Extract only nuclear power plants
        nuclear_ref = self.ref_dispatch[nuclear_names]
        nuclear_syn = self.syn_dispatch[nuclear_names]

        # Plot results
        fig, axes = plt.subplots(1, 2, figsize=(14, 5))
        axes[0].hist(nuclear_ref.values, bins=100)
        axes[1].hist(nuclear_syn.values, bins=100)
        axes[0].set_title('Nuclear Reference Distribution')
        axes[1].set_title('Nuclear Synthetic Distribution')

        if save_plots:
            # Save plot as pnd
            extent0 = axes[0].get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            extent1 = axes[1].get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            fig.savefig('images/nuclear_kpi/ref_dispatch_histogram.png', bbox_inches=extent0.expanded(1.3, 1.3))
            fig.savefig('images/nuclear_kpi/syn_dispatch_histogram.png', bbox_inches=extent1.expanded(1.3, 1.3))

        return None



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

from .PypsaDispatchBackend.EDispatch_L2RPN2020 import RampMode # TODO: Supprimer cette dépendance car pas utile (utiliser utils dans chronix2grid)
from .dispatch_loss_utils import run_grid2op_simulation_donothing, correct_scenario_loss, move_chronics_temporarily, \
    remove_temporary_chronics, remove_simulation_data, move_env_temporarily
import shutil
import os
import pathlib


def main(dispatcher, input_folder, output_folder, grid_folder, seed, params, params_opf,renewable_in_OPF=False):
    """

    Parameters
    ----------
    dispatcher : Dispatcher
        The Dispatcher instance used for running the OPF
    input_folder : str
        The path to the directory containing the inputs for the dispatch
    output_folder
        The path of the directory that will receive the outputs of the dispatch
    grid_folder
        The path of the directory that contains grid information (used only if loss simulation are activated)
    seed : int
        Random seed for parallel execution
    params_opf : dict
        Options for the OPF

    Returns
    -------
    DispatchResults
        The namedtuple return by Dispatcher.run method
    """

    #np.random.seed(seed) # already done before
    if ("renewable_in_opf" in params.keys()):
        renewable_in_OPF=params["renewable_in_opf"]
    else:
        renewable_in_OPF=False
    hydro_constraints = dispatcher.make_hydro_constraints_from_res_load_scenario()

    if renewable_in_OPF:
        load_with_losses = dispatcher.net_load(params_opf['losses_pct'],
                                               name=dispatcher.loads.index[0],include_renewable=False)
        ##########
        #Bypass solar and wind for now
        dispatch_results = dispatcher.run(
            load=load_with_losses,
            total_solar=dispatcher.solar_p.sum(axis=1),
            total_wind=dispatcher.wind_p.sum(axis=1),
            params=params_opf,
            gen_constraints=hydro_constraints,
            ramp_mode=parse_ramp_mode(params_opf['ramp_mode']),
            by_carrier=params_opf['dispatch_by_carrier'],
            pyomo=params_opf['pyomo'],
            solver_name=params_opf['solver_name']
        )
        #dispatch_results.chronix.prods_dispatch=dispatch_results.chronix.prods_dispatch.drop(["agg_wind", "agg_solar"], axis=1)
    else:
        load_with_losses = dispatcher.net_load(params_opf['losses_pct'],
                                               name=dispatcher.loads.index[0], include_renewable=True)
        ##########
        # Bypass solar and wind for now
        dispatch_results = dispatcher.run(
            load=load_with_losses,
            total_solar=None,
            total_wind=None,
            params=params_opf,
            gen_constraints=hydro_constraints,
            ramp_mode=parse_ramp_mode(params_opf['ramp_mode']),
            by_carrier=params_opf['dispatch_by_carrier'],
            pyomo=params_opf['pyomo'],
            solver_name=params_opf['solver_name']
        )
        #####
        # These column should be removed as wind and solar were not considered in the opf here
        dispatch_results.chronix.prods_dispatch = dispatch_results.chronix.prods_dispatch.drop(
            ["agg_wind", "agg_solar"], axis=1)

    dispatcher.save_results(params, output_folder)

    is_dispatch_successful=(dispatcher.chronix_scenario.prods_dispatch is not None) and (len(dispatcher.chronix_scenario.prods_dispatch.columns)>=1)
    if params_opf["loss_grid2op_simulation"] and is_dispatch_successful:
        new_prod_p, new_prod_forecasted_p = simulate_loss(grid_folder, output_folder, params_opf, write_results = True)
        dispatch_results = update_results_loss(dispatch_results, new_prod_p, params_opf)

    return dispatch_results

def update_results_loss(dispatch_results, new_prod_p, params_opf):
    dispatch_results[0].prods_dispatch[params_opf['nameSlack']] = new_prod_p[params_opf['nameSlack']]
    return dispatch_results

def simulate_loss(input_folder, output_folder, params_opf, write_results = True):
    scenario_folder_path = output_folder
    grid_folder_g2op = input_folder

    #because of possible multiprocessing, me need to create a new grid2op grid folder for each scenario
    #to avoid interferences
    grid_temporary_path=move_env_temporarily(scenario_folder_path, grid_folder_g2op)

    move_chronics_temporarily(scenario_folder_path, grid_temporary_path)
    agent_results_path = str(pathlib.Path(scenario_folder_path).parent.parent)
    # try:

    episode_data = run_grid2op_simulation_donothing(grid_temporary_path, scenario_folder_path,write_results=write_results,agent_results_path=agent_results_path)
    # except RuntimeError:
    #     remove_temporary_chronics(grid_folder_g2op)
    #     raise RuntimeError("Error in Grid2op simulation, temporary folder deleted")
    dispatch_results_corrected = correct_scenario_loss(scenario_folder_path, params_opf, grid_folder_g2op, episode_data)

    #remove temporary env folder
    shutil.rmtree(grid_temporary_path)

    # remove_simulation_data(scenario_folder_path)
    return dispatch_results_corrected


def parse_ramp_mode(mode):
    """
    Parse a string representing the difficulty of the ramps in the OPF into
    a RampMode Enum value
    Parameters
    ----------
    mode : str
        The difficulty mode for the ramps in the OPF

    Returns
    -------
    RampMode
        The encoded RampMode value

    """
    if mode == 'hard':
        return RampMode.hard
    if mode == 'medium':
        return RampMode.medium
    if mode == 'easy':
        return RampMode.easy
    if mode == '':
        return RampMode.none
    raise ValueError(f'mode only takes values from (hard, medium, easy, none), '
                     '{mode} was passed')



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

from .generate_load import main


class ConsumptionGeneratorBackend:
    """
    Backend that generates load power chronics that represent consumption nodes in the grid, with a spatiotemporal correlated noise model.
    It takes into account specific parameters about the grid, the solar and wind power plants, the regions settings...
    See in documentation *Description of implemented models* for detailed information about the model

    Attributes
    ----------
    out_path: ``str``
        path to output folder for generated chronics
    seed: ``int``
    params: ``dict``
        dictionnary with the model parameters. It needs to contain keys **"dt", "planned_std", "Lx", "Ly", "dx_corr", "dy_corr",
        "temperature_corr", "std_temperature_noise"**
    loads_charac: :class:`pandas.DataFrame`
        data frame with characteristics on load nodes in the simulated grid
    load_config_manager: :class:`chronix2grid.config.LoadsConfigManager`
        config manager used to load specific patterns used for the model (load weekly pattern for residential power consumption)
    write_results: ``bool``
    """
    def __init__(self, out_path, seed, params, loads_charac, load_config_manager, write_results, day_lag=0):
        self.write_results = write_results
        self.load_config_manager = load_config_manager
        self.loads_charac = loads_charac
        self.params = params
        self.seed = seed
        self.out_path = out_path
        self.day_lag = day_lag

    def run(self, load_weekly_pattern=None):
        """
        Runs the generation model in ``chronix2grid.generation.consumption.generate_load`` and writes chronics
        """
        if load_weekly_pattern is None:
            load_weekly_pattern = self.load_config_manager.read_specific()
        return main(self.out_path, self.seed, self.params, self.loads_charac, load_weekly_pattern, self.write_results,
                    day_lag=self.day_lag)



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import argparse
import numpy as np
import pandas as pd
import seaborn as sns
from itertools import product
from pathlib import Path
from matplotlib import pyplot as plt


class SolarKPI:

    def __init__(self, kpi_databanks):
        self.kpi_databanks = kpi_databanks

    def __solar_at_night(self,
                         solar_df,
                         params,
                         aggregated=False
                         ):

        '''
        '''

        if aggregated:
            solar_df = solar_df.sum(axis=1)

        # Extract parameters
        monthly_pattern = params['monthly_pattern']
        hours = params['hours']

        # Add month variable solar df
        solar_df_month = pd.concat([solar_df, self.months], axis=1)

        # Iterate over all season per month to check if solar is
        # present during night hours defined by a deterministic
        # criteria per season.
        season_at_night = {}
        for season, month in monthly_pattern.items():

            # Filter by season (including all possible months avaible in season)
            month_filter = solar_df_month.month.isin(month)
            season_solar = solar_df_month.loc[month_filter, solar_df_month.columns != 'month']

            if not season_solar.empty:
                # Filter by season and hours
                sum_over_season = season_solar.drop(season_solar.between_time(hours[season][0], hours[season][1]).index).sum(axis=0)
                percen_over_season = 100 * sum_over_season / season_solar.sum(axis=0)
                percen_over_season = percen_over_season.round(self.precision)

                # Save results in a dictionary
                season_at_night.update({season: percen_over_season})

        return season_at_night

    def __solar_cloudiness(self,
                           solar_df,
                           cloud_quantile,
                           factor_cloud):

        '''
        '''

        # Using all data, we get a unique x quantile
        # in order to grab a long term value
        solar_q = solar_df.quantile(cloud_quantile)

        # Per day, we are interested to get x quantile only when
        # generators are producing energy
        solar_q_perday = solar_df.replace(0, np.nan).resample('D').quantile(cloud_quantile)

        # Measure cloudiness: we compare the quantile values per
        # day with a long them x quantile truncated it a factor
        cloudiness = solar_q_perday <= (solar_q * factor_cloud)

        # # Add month column to get some particular stats
        # # Add months to solar cloudiness measure df
        # month = cloudiness.index.month.to_frame()
        # month.index = cloudiness.index
        # cloudiness['month'] = month

        # Get in percentage the number of days solar
        # generators have been producing below the factor
        # (We considerer as a cloudiness's day)
        # percen_cloud = 100 * cloudiness.groupby('month').drop('month', axis=1).sum() \
        #                    / cloudiness.groupby('month').drop('month', axis=1).count()
        # percen_cloud = percen_cloud.round(self.precision)

        percen_cloud = 100 * cloudiness.groupby(pd.Grouper(freq='M')).sum() \
                       / cloudiness.groupby(pd.Grouper(freq='M')).count()

        percen_cloud.index = percen_cloud.index.month
        percen_cloud.index.rename('month', inplace=True)

        return percen_cloud.round(self.precision)

    def solar_kpi(self,
                  cloud_quantile=0.95,
                  cond_below_cloud=0.57,
                  **kwargs):

        '''

        Parameters:
        ----------

        moderate_cloud_quantile: Quantile applied to the consequitive
                                 difference solar hisrogram to determine
                                 large spiked due to cloudness.
        severe_cloud_quantile:   Quantile applied to the consequitive
                                 difference solar hisrogram to determine
                                 large spiked due to cloudness.

        **kwargs:

            monthly_pattern: Contains a dictionary to define the month
                                a solar time series shoud follow (e.g summer
                                months should have more production rather than
                                sping).

                        monthly_pattern = {'summer': [6,7,8],
                                           'fall': [9,10,11],
                                           'winter': [12,1,2],
                                           'spring': [2,3,4,5]}

            hours: Defines the hours solar is producing energy to the system
                        per season.

                        hours = {'summer': ('07:00', '20:00'),
                                 'fall': ('08:00', '18:00'),
                                 'winter': ('09:30', '16:30')}

        Returns:
        --------

        corr_solar (pd.DataFrame):  Correlation matrix between reference and synthetic data

        season_night_percen (dict): Dictionary containing season percentage solar units has been
                                    producing energy out of hours.

        percen_moderate (dict):     Dictionary per solar unit that contains the number of times in
                                    percentage and per month a specific solar unit has overcome a pre-defined
                                    moderate qunantile given the ref dispatch.

        percen_severe (dict):       Dictionary per solar unit that contains the number of times in
                                    percentage and per month a specific solar unit has overcome a pre-defined
                                    severe qunantile given the ref dispatch.

        '''

        # Get the solar gen names
        solar_filter = self.prod_charac.type.isin(['solar'])
        solar_names = self.prod_charac.name.loc[solar_filter].values

        # From data, extract only wind time series
        solar_ref = self.ref_dispatch[solar_names]
        solar_syn = self.syn_dispatch[solar_names]

        # First KPI
        # -- + -- +
        # Get correlation matrix (10 x 10)
        corr_solar = self._pairwise_corr_different_dfs(solar_ref, solar_syn)

        # Write its value
        # -- + -- + -- +
        self.output['solar_kpi'] = {'solar_corr': corr_solar.to_dict()}

        # Second KPI
        # -- + -- +
        # Get percentage solar at night

        if not kwargs:
            monthly_pattern = {'summer': [6, 7, 8], 'fall': [9, 10, 11],
                               'winter': [12, 1, 2], 'spring': [2, 3, 4, 5]}

            hours = {'summer': ('07:00', '20:00'),
                     'fall': ('08:00', '18:00'),
                     'winter': ('09:30', '16:30'),
                     'spring': ('08:00', '18:00')}

        # Get percentage solar productions for reference data
        params = {'monthly_pattern': monthly_pattern, 'hours': hours}

        solar_night_ref = self.__solar_at_night(solar_ref, params=params)

        # Get percentage solar productions for synthetic data
        solar_night_syn = self.__solar_at_night(solar_syn, params=params)

        # Write output
        # -- + -- + --
        self.output['solar_kpi'] = {'season_solar_at_night_reference': solar_night_ref,
                                    'season_solar_at_night_synthetic': solar_night_syn,
                                    }

        # Third KPI
        # -- + -- +
        # Measure Cloudiness as counting the number of days
        # per month when the x quantile is below a factor
        # specified by the method.

        cloudiness_ref = self.__solar_cloudiness(solar_ref,
                                                 cloud_quantile=cloud_quantile,
                                                 factor_cloud=cond_below_cloud)

        cloudiness_syn = self.__solar_cloudiness(solar_syn,
                                                 cloud_quantile=cloud_quantile,
                                                 factor_cloud=cond_below_cloud)

        # # Write its value
        # # -- + -- + -- +
        self.output['solar_kpi'] = {'cloudiness_reference': cloudiness_ref.to_dict(),
                                    'cloudiness_synthetic': cloudiness_syn.to_dict()
                                    }

        return corr_solar, solar_night_ref, solar_night_syn, cloudiness_ref, cloudiness_syn



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

from .generate_dispatch import main


class DispatchBackend: # TODO - PypsaDispatchBackend - devra créer un PypsaDispatcher et l'utiliser. OU ALORS dans les constantes de Chronix2grid choisir la classe de Dispatcher
    """
    Backend that generates nuclear, hydro and thermal from consumption and renewable production constraints by
    computing an optimized economic dispatch.
    Constraints on generators and consumption, solar, wind and loss chronics are provided
    via an instance of :class:`chronix2grid.dispatch.EconomicDispatch.Dispatcher` object

    Attributes
    ----------
    dispatcher: :class:`chronix2grid.dispatch.EconomicDispatch.Dispatcher`
        Object that contains grid and chronics features that represents the dispatch constraints and environment
    scenario_folder_path: ``str``
        path to output folder for generated chronics
    grid_folder: ``str``
        folder in which grid info are stored
    seed_disp: ``int``
    params: ``dict``
        dictionnary with the model parameters. It needs to contain keys **"dt", "planned_std"**
    params_opf: ``dict``
        dictionnary with specific parameters concerning the dispatch optimization (Optimal Power Flow computation)
    """
    def __init__(self,
                 dispatcher,
                 scenario_folder_path,
                 grid_folder,
                 seed_disp,
                 params,
                 params_opf):
        self.dispatcher = dispatcher
        self.params = params
        self.params_opf = params_opf
        self.seed_disp = seed_disp
        self.scenario_folder_path = scenario_folder_path
        self.grid_folder = grid_folder

    def run(self):
        """
        Runs Economic dispatch as in ``chronix2grid.dispatch.generate_dispatch``

        .. warning::
            The dispatch optimization can rely on pypsa simulation. If it is the case you should ensure pypsa dependencies are installed

        .. note::
            As a final step, loss can be simulated thanks to grid2op. It is achieved if "loss_grid2op_simulation" is True in params_opf.
            You should then provide keys **"idxSlack","nameSlack","early_stopping_mode","pmin_margin","pmax_margin","rampup_margin",
            "rampdown_margin","agent_type"**

        """
        return main(self.dispatcher, self.scenario_folder_path, self.scenario_folder_path,
                    self.grid_folder, self.seed_disp, self.params, self.params_opf)



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

"""Class for the economic dispatch framework. Allows to parametrize and run
an economic dispatch based on RES and consumption time series"""

import copy
import os
from abc import ABC, abstractmethod
from copy import deepcopy
import datetime as dt
from collections import namedtuple
import pathlib
from numpy.random import default_rng

import grid2op
from grid2op.Chronics import ChangeNothing
import pandas as pd
import plotly.express as px

from chronix2grid.generation.dispatch.utils import RampMode, add_noise_gen, modify_hydro_ramps, modify_slack_characs
import chronix2grid.constants as cst

DispatchResults = namedtuple('DispatchResults', ['chronix', 'terminal_conditions'])

def init_dispatcher_from_config(env_path, input_folder, dispatcher_class, params_opf):
    # Read grid and gens characs
    env118_withoutchron = grid2op.make(env_path,
                                       test=True,
                                       chronics_class=ChangeNothing)
    # grid_path_parent = pathlib.Path(grid_path).parent.absolute()
    # env118_withoutchron = grid2op.make(str(grid_path_parent),
    #                                 chronics_path=str(grid_path_parent),
    #                                 chronics_class=ChangeNothing)
    # env118_withoutchron = grid2op.make(grid_path)

    # Dispatcher object init
    dispatcher = dispatcher_class.from_gri2op_env(env118_withoutchron)

    dispatcher.read_hydro_guide_curves(
        os.path.join(input_folder, 'patterns', 'hydro_french.csv'))
    return dispatcher

def init_dispatcher_from_config_dataframe(grid_path, input_folder, dispatcher_class, params_opf):
    # Read grid and gens characs
    # env118_withoutchron = grid2op.make("rte_case118_example",
    #                                    test=True,
    #                                    grid_path=grid_path,
    #                                    chronics_class=ChangeNothing)

    # Put infos in dataframe
    # env_df = pd.DataFrame({'name':env118_withoutchron.name_gen,
    #                        'type':env118_withoutchron.gen_type,
    #                        'pmax':env118_withoutchron.gen_pmax,
    #                        'max_ramp_up':env118_withoutchron.gen_max_ramp_up,
    #                        'max_ramp_down':env118_withoutchron.gen_max_ramp_down,
    #                        'cost_per_mw':env118_withoutchron.gen_cost_per_MW})

    # Reading characs from file directly avoids some strange problem with grid2op when launching Chronix2grid several times in the same session
    prods_charac = pd.read_csv(os.path.join(pathlib.Path(grid_path).parent.absolute(),"prods_charac.csv"), decimal = '.')
    env_df = prods_charac.rename(columns = {"Pmax":'pmax',
                                            "marginal_cost":"cost_per_mw"})[["name","type","pmax","max_ramp_up",
                                                                             "max_ramp_down","cost_per_mw"]]

    # Generators temporary adjusts
    env_df = modify_hydro_ramps(env_df, params_opf["hydro_ramp_reduction_factor"])

    if params_opf["slack_p_max_reduction"] != 0. or params_opf['slack_ramp_max_reduction'] != 0.:
        env_df = modify_slack_characs(env_df,
                                                 params_opf["nameSlack"],
                                                 params_opf["slack_p_max_reduction"],
                                                 params_opf["slack_ramp_max_reduction"])

    # Dispatcher object init
    dispatcher = dispatcher_class.from_dataframe(env_df)

    dispatcher.read_hydro_guide_curves(
        os.path.join(input_folder, 'patterns', 'hydro_french.csv'))
    return dispatcher



class Dispatcher(ABC):
    """Abstract Class Dispatcher with some operational method that don't depend on the dispatcher technology (pypsa...)
    and some others that are implemented specifically in inheriting classes"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def run(self, load, params, gen_constraints=None,
            ramp_mode=RampMode.hard, by_carrier=False, **kwargs):
        """Run the proper dispatch optimization. Have to be implemented in inheriting classes"""

    @abstractmethod
    def simplify_net(self):
        pass

    @classmethod
    @abstractmethod
    def from_gri2op_env(cls, grid2op_env):
        """Reads grid features from a grid2op environment into a specific object.
        Have to be implemented in inheriting classes according to the type of model"""

    @classmethod
    @abstractmethod
    def from_dataframe(cls, env_df):
        """Reads grid features from a pandas DataFrame into a specific object.
        Have to be implemented in inheriting classes according to the type of model"""

    @property
    def wind_p(self):
        if self._chronix_scenario is None:
            raise Exception('Cannot access this property before instantiated the Load and '
                            'renewables scenario.')
        return self._chronix_scenario.wind_p

    @property
    def solar_p(self):
        if self._chronix_scenario is None:
            raise Exception('Cannot access this property before instantiated the Load and '
                            'renewables scenario.')
        return self._chronix_scenario.solar_p

    @property
    def chronix_scenario(self):
        """

        Returns
        -------
        chronix_scenario: :class:`chronix2grid.generation.dispatch.EconomicDispatch.ChroniXScenario`
            Object that contains all the chronics that are relevant to perform a dispatch
        """
        if self._chronix_scenario is None:
            raise Exception('Cannot access this property before instantiated the Load and '
                            'renewables scenario.')
        return self._chronix_scenario

    @chronix_scenario.setter
    def chronix_scenario(self, chronix_scenario):
        if not isinstance(chronix_scenario, ChroniXScenario):
            raise Exception('The chronix_scenario argument should be an instance of '
                            'ChronixScenario.')
        self._chronix_scenario = chronix_scenario

    def net_load(self, losses_pct, name,include_renewable=True):
        if self._chronix_scenario is None:
            raise Exception('Cannot compute net load before instantiated the Load and'
                            'renewables scenario.')
        return self._chronix_scenario.net_load(losses_pct, name)

    def nlargest_ramps(self, n, losses_pct):
        ramps = self.net_load(losses_pct, "").diff()
        return ramps.nlargest(n, ramps.columns[0])

    def reset_ramps_from_grid2op_env(self):
        if self._env is None:
            raise Exception('This method can only be applied when Dispatch has been'
                            'instantiated from a grid2op Environment.')
        for i, generator in enumerate(self._env.name_gen):
            if generator in self.generators.index:
                self.generators.loc[generator, 'ramp_limit_up'] = \
                    self._env.gen_max_ramp_up[i] / self._env.gen_pmax[i]
                self.generators.loc[generator, 'ramp_limit_down'] = \
                    self._env.gen_max_ramp_down[i] / self._env.gen_pmax[i]

    def reset_ramps_from_dataframe(self):
        if self._df is None:
            raise Exception('This method can only be applied when Dispatch has been'
                            'instantiated from a pandas DataFrame.')
        for i, (generator,pmax, rampup, rampdown) in enumerate(zip(self._df['name'],
                                                                   self._df['pmax'],
                                                                   self._df['max_ramp_up'],
                                                                   self._df['max_ramp_down'])):
            if generator in self.generators.index:
                self.generators.loc[generator, 'ramp_limit_up'] = \
                    rampup / pmax
                self.generators.loc[generator, 'ramp_limit_down'] = \
                    rampdown / pmax

    def read_hydro_guide_curves(self, hydro_file_path):
        """
        Reads realistic hydro pattern that provides seasonal boundaries to the hydro production.
        This constraint in the dispatch problem leads to more realistic hydro production

        Parameters
        ----------
        hydro_file_path: ``str``

        """
        dateparse = lambda x: dt.datetime.strptime(x, '%Y-%m-%d %H:%M')
        hydro_pattern = pd.read_csv(hydro_file_path, usecols=[0, 2, 3],
                                    parse_dates=[0], date_parser=dateparse)
        hydro_pattern.set_index(hydro_pattern.columns[0], inplace=True)
        hydro_names = self.generators[self.generators.carrier == 'hydro'].index

        for extremum in ['min', 'max']:
            hydro_pu = hydro_pattern[[f'p_{extremum}_u'] * len(hydro_names)]
            hydro_pu.columns = hydro_names
            hydro_pu.index = hydro_pu.index.map(
                lambda x: (x.month, x.day, x.hour, x.minute, x.second))
            setattr(self, f'_{extremum}_hydro_pu', hydro_pu)

        self._hydro_file_path = hydro_file_path

    def read_load_and_res_scenario(self, load_path_file, prod_path_file,
                                   scenario_name, start_date, end_date, dt):
        if self._env is None:
            raise Exception('This method can only be applied when Dispatch has been'
                            'instantiated from a grid2op Environment.')
        res_names = dict(
            wind=[name for i, name in enumerate(self._env.name_gen)
                  if self._env.gen_type[i] == 'wind'],
            solar=[name for i, name in enumerate(self._env.name_gen)
                   if self._env.gen_type[i] == 'solar']
        )
        self._chronix_scenario = ChroniXScenario.from_disk(
            load_path_file, prod_path_file,
            res_names, scenario_name=scenario_name,
            start_date=start_date, end_date=end_date, dt=dt
        )


    def make_hydro_constraints_from_res_load_scenario(self):
        if self._chronix_scenario is None or self._hydro_file_path is None:
            raise Exception('This method can only be applied when a Scenario for load'
                            'and renewables has been instantiated and hydro guide'
                            'curves have been read.')
        index_slice = self._chronix_scenario.loads.index.map(
            lambda x: (x.month, x.day, x.hour, x.minute, x.second)
        )

        self._min_hydro_pu = self._min_hydro_pu.reindex(index_slice).fillna(method='ffill')
        self._max_hydro_pu = self._max_hydro_pu.reindex(index_slice).fillna(method='ffill')

        return {'p_max_pu': self._max_hydro_pu.copy(), 'p_min_pu': self._min_hydro_pu.copy()}

    def modify_marginal_costs(self, new_costs):
        for carrier, new_cost in new_costs.items():
            try:
                targeted_generators = self.generators.carrier.isin([carrier])
                self.generators.loc[targeted_generators, 'marginal_cost'] = new_cost
            except KeyError:
                print(f'Carrier {carrier} is not a valid carrier')
                raise

    def plot_ramps(self):
        caract_gen = self.generators[['p_nom', 'carrier', 'ramp_limit_up']].reset_index()
        caract_gen = caract_gen.rename(columns={'index': 'name'})

        fig = px.scatter(
            caract_gen, x='p_nom', y='ramp_limit_up', color='carrier',
            hover_data=['name']
        )
        return fig

    def save_results(self, params, output_folder, prng=None):
        """
        Saves dispatch results in prod_p.csv.bz2, prod_p_forecasted.csv.bz2, load_p.csv.bz2, prices.csv

        Parameters
        ----------
        params: ``dict``
        output_folder: ``str``

        """
        if prng is None:
            prng = default_rng()
        if not self._has_results and not self._has_simplified_results:
            print('The optimization has first to run successfully in order to '
                  'save results.')
        if self._has_results:
            print('Saving results for the grids with individual generators...')
            res_load_scenario = self.chronix_scenario
        else:
            print('Saving results for the grids with aggregated generators by carriers...')
            res_load_scenario = self._simplified_chronix_scenario

        path_metadata_failed = os.path.join(output_folder, "DISPATCH_FAILED")
        if res_load_scenario is None:
            # the backend failed to find a solution
            print('ERROR: the backend failed to find a consistent state. Nothing is saved.')
            with open(path_metadata_failed, "w", encoding="utf-8") as f:
                f.write("The dispatch has failed. We cannot do anything.")
            return
        
        # this did not failed, so I remove it
        if os.path.exists(path_metadata_failed):
            os.remove(path_metadata_failed)
        
        wind_curtail_coeff = 1.0
        solar_curtail_coeff = 1.0
        
        # TODO perf do not recompute the res_load_scenario.wind_p.sum(axis=1)
        if "agg_wind" in res_load_scenario.prods_dispatch:
            wind_curtail_coeff = res_load_scenario.prods_dispatch["agg_wind"] / res_load_scenario.wind_p.sum(axis=1)
            wind_curtail_coeff = wind_curtail_coeff.values.reshape(-1, 1)
        if "agg_solar" in res_load_scenario.prods_dispatch:
            solar_curtail_coeff = res_load_scenario.prods_dispatch["agg_solar"] / res_load_scenario.solar_p.sum(axis=1)
            solar_curtail_coeff = solar_curtail_coeff.values.reshape(-1, 1)
        
        new_wind_after_curtail = res_load_scenario.wind_p * wind_curtail_coeff
        new_solar_after_curtail = res_load_scenario.solar_p * solar_curtail_coeff
        
        full_opf_dispatch = pd.concat(
            [res_load_scenario.prods_dispatch,
             new_wind_after_curtail,
             new_solar_after_curtail
            ],
            axis=1
        )
        
        diff_wind = (res_load_scenario.wind_p - new_wind_after_curtail).sum(axis=1).values
        total_curt = diff_wind.sum()
        total_wind = res_load_scenario.wind_p.sum(axis=1).sum()
        print(f"INFO: wind curtailment max: {diff_wind.max():.2f}MW")
        print(f"INFO: wind curtailment min: {diff_wind.min():.2f}MW")
        print(f"INFO: wind curtailment sum: {total_curt / 12.:.2f}MWh (total {total_wind / 12.:.2f}MWh: {100. * total_curt / total_wind:.2f}%)")
        diff_solar = (res_load_scenario.solar_p - new_solar_after_curtail).sum(axis=1).values
        total_curt = diff_solar.sum()
        total_solar = res_load_scenario.solar_p.sum(axis=1).sum()
        print(f"INFO: solar curtailment max: {diff_solar.max():.2f}MW")
        print(f"INFO: solar curtailment min: {diff_solar.min():.2f}MW")
        print(f"INFO: solar curtailment sum: {total_curt / 12.:.2f}MWh (total { total_curt / total_solar:.2f}MWh: {100. * total_curt / total_solar :.2f}%)")
        
        try:
            if self._env is not None:
                full_opf_dispatch = full_opf_dispatch[self._env.name_gen].round(2)
            else:
                full_opf_dispatch = full_opf_dispatch[self._df['name'].values].round(2)
        except KeyError:
            # Either we're trying to save results from a simplified dispatch or
            # using the save function before instanciating an env.
            pass
        if self._env is not None:
            gen_cap = pd.Series({gen_name: gen_pmax for gen_name, gen_pmax in
                                 zip(self._env.name_gen, self._env.gen_pmax)})
        else:
            gen_cap = pd.Series({gen_name: gen_pmax for gen_name, gen_pmax in
                                 zip(self._df['name'], self._df['pmax'])})

        prod_p_forecasted_with_noise = add_noise_gen(prng,
                                                     copy.deepcopy(full_opf_dispatch),
                                                     gen_cap,
                                                     noise_factor=params['planned_std'])

        # prod_p_forecasted_with_noise.to_csv(
        prod_p_forecasted_with_noise.to_csv(
            os.path.join(output_folder, "prod_p_forecasted.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )
        # prod_p_with_noise.to_csv(
        full_opf_dispatch.to_csv(
            os.path.join(output_folder, "prod_p.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )
        res_load_scenario.marginal_prices.to_csv(
            os.path.join(output_folder, "prices.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )
        res_load_scenario.loads.to_csv(
            os.path.join(output_folder, "load_p.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )
        # save the origin time series        
        pd.concat([res_load_scenario.wind_p, res_load_scenario.solar_p], axis=1).to_csv(
            os.path.join(output_folder, "prod_p_renew_orig.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )

class ChroniXScenario:
    def __init__(self, loads, prods, res_names, scenario_name, loss=None):
        self.loads = loads
        self.wind_p = prods[res_names['wind']]
        self.solar_p = prods[res_names['solar']]
        self.total_res = pd.concat([self.wind_p, self.solar_p], axis=1).sum(axis=1)
        self.prods_dispatch = None  # Will receive the results of the dispatch
        self.marginal_prices = None  # Will receive the marginal prices associated to a dispatch
        self.name = scenario_name
        self.loss = loss

    @classmethod
    def from_disk(cls, load_path_file, prod_path_file, res_names, scenario_name,
                  start_date, end_date, dt, loss_path_file=None):
        loads = pd.read_csv(load_path_file, sep=';')
        prods = pd.read_csv(prod_path_file, sep=';')
        if loss_path_file is not None:
            loss = pd.read_csv(loss_path_file, sep=';')
        else:
            loss = None
        datetime_index = pd.date_range(
            start=start_date,
            end=end_date,
            freq=str(dt) + 'min')
        loads.index = datetime_index[:len(loads)]
        prods.index = datetime_index[:len(prods)]
        return cls(loads, prods, res_names, scenario_name, loss)

    def net_load(self, losses_pct, name,include_renewable=True):
        if self.loss is None:
            load_minus_losses = self.loads.sum(axis=1) * (1 + losses_pct / 100)
        else:
            load_minus_losses = self.loads.sum(axis=1) + self.loss
        if include_renewable:
            return (load_minus_losses - self.total_res).to_frame(name=name)
        else:
            return (load_minus_losses).to_frame(name=name)

    def simplify_chronix(self):
        simplified_chronix = deepcopy(self)
        simplified_chronix.wind_p = simplified_chronix.wind_p.sum(axis=1).to_frame(name='wind')
        simplified_chronix.solar_p = simplified_chronix.solar_p.sum(axis=1).to_frame(name='solar')
        simplified_chronix.name = simplified_chronix.name + 'by_carrier'
        return simplified_chronix


# if __name__ == "__main__":
#
#     INPUT_FOLDER = 'chronix2grid/generation/input'
#     CASE = 'case118_l2rpn'
#     path_grid = os.path.join(INPUT_FOLDER, CASE)
#
#     losses_pct = 3.0
#
#     env118_blank = grid2op.make(
#         test=True,
#         grid_path=os.path.join(path_grid, "L2RPN_2020_case118_redesigned.json"),
#         chronics_class=ChangeNothing,
#     )
#     params = {'snapshots': [],
#               'step_opf_min': 5,
#               'mode_opf': 'week',
#               'reactive_comp': 1.025,
#               }
#     chronics_path_gen = os.path.join(INPUT_FOLDER, "dispatch", str(2012))
#     this_path = os.path.join(chronics_path_gen, 'Scenario_0')
#     dispatch = Dispatcher.from_gri2op_env(env118_blank)
#     dispatch.read_hydro_guide_curves(os.path.join(INPUT_FOLDER, 'patterns', 'hydro.csv'))
#     dispatch.read_load_and_res_scenario(os.path.join(this_path, 'load_p.csv.bz2'),
#                                         os.path.join(this_path, 'prod_p.csv.bz2'),
#                                         'Scenario_0')
#     dispatch.make_hydro_constraints_from_res_load_scenario()
#     net_by_carrier = dispatch.simplify_net()
#     agg_load_without_renew = dispatch.net_load(losses_pct, name=dispatch.loads.index[0])
#
#     # Prepare gen constraints for EDispatch module
#     hydro_constraints = {'p_max_pu': dispatch._max_hydro_pu.copy(),
#                          'p_min_pu': dispatch._min_hydro_pu.copy()}
#
#     opf_dispatch, term_conditions = dispatch.run(
#         agg_load_without_renew,
#         params=params,
#         gen_constraints=hydro_constraints,
#         ramp_mode=run_economic_dispatch.RampMode.easy,
#         by_carrier=False  # True to run the dispatch only aggregated generators by carrier
#     )
#
#     dispatch.save_results(params, '.')
#     test_prods = pd.read_csv('./Scenario_0/prod_p.csv.bz2', sep=";")
#     test_prices = pd.read_csv('./Scenario_0/prices.csv.bz2', sep=";")



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

from .generate_solar_wind import main


class RenewableBackend:
    """
    Backend that generates solar and wind production chronics with a spatiotemporal correlated noise model.
    It takes into account specific parameters about the grid, the solar and wind power plants, the regional setting...
    See in documentation *Description of implemented models* for detailed information about the model

    Attributes
    ----------
    out_path: ``str``
        path to output folder for generated chronics
    seed: ``int``
    params: ``dict``
        dictionnary with the model parameters. It needs to contain keys **"dt", "planned_std", "Lx", "Ly", "dx_corr", "dy_corr", "short_wind_corr",
        "medium_wind_corr", "long_wind_corr", "solar_corr", "smoothdist", "std_short_wind_noise", "std_solar_noise",
        "std_medium_wind_noise", "std_long_wind_noise", "year_solar_pattern"**
    prods_charac: :class:`pandas.DataFrame`
        data frame with characteristics on wind and solar generators/power plants of the model
    res_config_manager: :class:`chronix2grid.config.ResConfigManager`
        config manager used to load specific patterns used for the model (solar_pattern)
    write_results: ``bool``
    """
    def __init__(self, out_path, seed, params, loads_charac, res_config_manager, write_results):
        self.write_results = write_results
        self.res_config_manager = res_config_manager
        self.loads_charac = loads_charac
        self.params = params
        self.seed = seed
        self.out_path = out_path

    def run(self, solar_pattern=None):
        """
        Runs the generation model in ``chronix2grid.generation.renewable.generate_solar_wind`` and writes chronics
        """
        if solar_pattern is None:
            solar_pattern = self.res_config_manager.read_specific()
        return main(self.out_path, self.seed, self.params, self.loads_charac, solar_pattern, self.write_results)



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import json

# Other Python libraries
import pandas as pd
import numpy as np
from numpy.random import default_rng

# Libraries developed for this module
from . import solar_wind_utils as swutils
from .. import generation_utils as utils
import chronix2grid.constants as cst


def main(scenario_destination_path, seed, params, prods_charac, solar_pattern, write_results = True):
    """
    This is the solar and wind production generation function, it allows you to generate consumption chronics based on
    production nodes characteristics and on a solar typical yearly production patterns.

    Parameters
    ----------
    scenario_destination_path (str): Path of output directory
    seed (int): random seed of the scenario
    params (dict): system params such as timestep or mesh characteristics
    prods_charac (pandas.DataFrame): characteristics of production nodes such as Pmax and type of production
    solar_pattern (pandas.DataFrame): hourly solar production pattern for a year. It represent specificity of the production region considered
    smoothdist (float): parameter for smoothing
    write_results (boolean): whether to write results or not. Default is True

    Returns
    -------
    pandas.DataFrame: solar production chronics generated at every node with additional gaussian noise
    pandas.DataFrame: solar production chronics forecasted for the scenario without additional gaussian noise
    pandas.DataFrame: wind production chronics generated at every node with additional gaussian noise
    pandas.DataFrame: wind production chronics forecasted for the scenario without additional gaussian noise
    """

    prng = default_rng(seed)
    #np.random.seed(seed) #olver version - to be removed
    smoothdist = params['smoothdist']

    # Define datetime indices
    datetime_index = pd.date_range(
        start=params['start_date'],
        end=params['end_date'],
        freq=str(params['dt']) + 'min')

    # Solar_pattern management
    # Extra value (resolution 1H, 8761)
    solar_pattern = solar_pattern[:-1]

    # Realistic first day of year: have to roll the pattern to fit first day of week
    # start_date = params['start_date']
    # start_date_day = start_date.weekday()
    # pattern_start_date = pd.Timestamp("01-01-"+str(int(params['year_solar_pattern'])))
    # pattern_start_date_day = pattern_start_date.weekday()
    # days_to_shift = start_date_day - pattern_start_date_day
    # steps_to_shift = int(days_to_shift * 60 * 24 / params['dt']) # Solar pattern starts on a monday at 0h + timestep
    # solar_pattern = np.roll(solar_pattern, steps_to_shift)

    # Generate GLOBAL temperature noise
    print('Computing global auto-correlated spatio-temporal noise for sun and wind...')
    scale_solar_coord_for_correlation = float(params["scale_solar_coord_for_correlation"]) if "scale_solar_coord_for_correlation" in params else None
    add_dim = 0
    dx_corr = int(params['dx_corr'])
    dy_corr = int(params['dy_corr'])
    for x, y, type_gen  in zip(prods_charac["x"], prods_charac["y"], prods_charac["type"]):
        if type_gen == "solar" and scale_solar_coord_for_correlation is not None:
            x = scale_solar_coord_for_correlation * x
            y = scale_solar_coord_for_correlation * y
        x_plus = int(x // dx_corr + 1)
        y_plus = int(y // dy_corr + 1)
        add_dim = max(y_plus, add_dim)
        add_dim = max(x_plus, add_dim)

    solar_noise = utils.generate_coarse_noise(prng, params, 'solar', add_dim=add_dim)
    long_scale_wind_noise = utils.generate_coarse_noise(prng, params, 'long_wind', add_dim=add_dim)
    medium_scale_wind_noise = utils.generate_coarse_noise(prng, params, 'medium_wind', add_dim=add_dim)
    short_scale_wind_noise = utils.generate_coarse_noise(prng, params, 'short_wind', add_dim=add_dim)

    # Compute Wind and solar series of scenario
    print('Generating solar and wind production chronics')
    prods_series = {}
    for name in prods_charac['name']:
        mask = (prods_charac['name'] == name)
        if prods_charac[mask]['type'].values == 'solar':
            locations = [prods_charac[mask]['x'].values[0], prods_charac[mask]['y'].values[0]]
            Pmax = prods_charac[mask]['Pmax'].values[0]
            prods_series[name] = swutils.compute_solar_series(
                prng,
                locations,
                Pmax,
                solar_noise,
                params, solar_pattern, smoothdist,
                time_scale=params['solar_corr'],
                add_dim=add_dim,
                scale_solar_coord_for_correlation=scale_solar_coord_for_correlation)

        elif prods_charac[mask]['type'].values == 'wind':
            locations = [prods_charac[mask]['x'].values[0], prods_charac[mask]['y'].values[0]]
            Pmax = prods_charac[mask]['Pmax'].values[0]
            prods_series[name] = swutils.compute_wind_series(
                prng,
                locations,
                Pmax,
                long_scale_wind_noise,
                medium_scale_wind_noise,
                short_scale_wind_noise,
                params, smoothdist,
                add_dim=add_dim)

    # Séparation ds séries solaires et éoliennes
    solar_series = {}
    wind_series = {}
    for name in prods_charac['name']:
        mask = (prods_charac['name'] == name)
        if prods_charac[mask]['type'].values == 'solar':
            solar_series[name] = prods_series[name]
        elif prods_charac[mask]['type'].values == 'wind':
            wind_series[name] = prods_series[name]

    # Time index
    prods_series['datetime'] = datetime_index
    solar_series['datetime'] = datetime_index
    wind_series['datetime'] = datetime_index

    # Save files
    if scenario_destination_path is not None:
        print('Saving files in zipped csv')
        if not os.path.exists(scenario_destination_path):
            os.makedirs(scenario_destination_path)
            
    prod_solar_forecasted =  swutils.create_csv(
        prng,
        solar_series,
        os.path.join(scenario_destination_path, 'solar_p_forecasted.csv.bz2') if scenario_destination_path is not None else None,
        reordering=True,
        shift=True,
        write_results=write_results,
        index=False
    )

    prod_solar = swutils.create_csv(
        prng,
        solar_series,
        os.path.join(scenario_destination_path, 'solar_p.csv.bz2') if scenario_destination_path is not None else None,
        reordering=True,
        noise=params['planned_std'],
        write_results=write_results
    )

    prod_wind_forecasted = swutils.create_csv(
        prng,
        wind_series,
        os.path.join(scenario_destination_path, 'wind_p_forecasted.csv.bz2') if scenario_destination_path is not None else None,
        reordering=True,
        shift=True,
        write_results=write_results,
        index=False
    )

    prod_wind = swutils.create_csv(
        prng,
        wind_series, os.path.join(scenario_destination_path, 'wind_p.csv.bz2') if scenario_destination_path is not None else None,
        reordering=True,
        noise=params['planned_std'],
        write_results=write_results
    )

    prod_p = swutils.create_csv(
        prng,
        prods_series, os.path.join(scenario_destination_path, 'prod_p.csv.bz2') if scenario_destination_path is not None else None,
        reordering=True,
        noise=params['planned_std'],
        write_results=write_results
    )

    prod_v = prods_charac[['name', 'V']].set_index('name')
    prod_v = prod_v.T
    prod_v.index = [0]
    prod_v = prod_v.reindex(range(len(prod_p)))
    prod_v = prod_v.fillna(method='ffill') * 1.04
    
    if write_results:
        prod_v.to_csv(
            os.path.join(scenario_destination_path, 'prod_v.csv.bz2') if scenario_destination_path is not None else None,
            sep=';',
            index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )

    return prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted


# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import json
from numpy.random import default_rng

# Other Python libraries
import pandas as pd
import numpy as np

# Libraries developed for this module
from . import consumption_utils as conso
from .. import generation_utils as utils


def main(scenario_destination_path, seed, params, loads_charac, load_weekly_pattern, write_results = True, day_lag=0):
    """
    This is the load generation function, it allows you to generate consumption chronics based on demand nodes characteristics and on weekly demand patterns.

    Parameters
    ----------
    scenario_destination_path (string): where results are written
    seed (int): random seed of the scenario
    params (dict): system params such as timestep or mesh characteristics
    loads_charac (pandas.DataFrame): characteristics of loads node such as Pmax and type of demand
    load_weekly_pattern (pandas.DataFrame): 5 minutes weekly load chronic that represent specificity of the demand context
    write_results (boolean): whether to write_results or not. Default is True

    Returns
    -------
    pandas.DataFrame: loads chronics generated at every node with additional gaussian noise
    pandas.DataFrame: loads chronics forecasted for the scenario without additional gaussian noise
    """

    # Set random seed of scenario
    prng = default_rng(seed)
    #np.random.seed(seed) #older version - to be removed

    # Define reference datetime indices
    datetime_index = pd.date_range(
        start=params['start_date'],
        end=params['end_date'],
        freq=str(params['dt']) + 'min')


    add_dim = 0
    dx_corr = int(params['dx_corr'])
    dy_corr = int(params['dy_corr'])
    for x,y  in zip(loads_charac["x"], loads_charac["y"]):
        x_plus = int(x // dx_corr + 1)
        y_plus = int(y // dy_corr + 1)
        add_dim = max(y_plus, add_dim)
        add_dim = max(x_plus, add_dim)
        #add_dim=0 #to get back to when this parameter did not exist - to be removed
    
    # Generate GLOBAL temperature noise
    print('Computing global auto-correlated spatio-temporal noise for thermosensible demand...') ## temperature is simply to reflect the fact that loads is correlated spatially, and so is the real "temperature". It is not the real temperature.
    temperature_noise = utils.generate_coarse_noise(prng, params, 'temperature', add_dim=add_dim)

    print('Computing loads ...')
    start_day = datetime_index[0]
    loads_series = conso.compute_loads(loads_charac,
                                       temperature_noise,
                                       params,
                                       load_weekly_pattern,
                                       start_day=start_day,
                                       add_dim=add_dim,
                                       day_lag=day_lag)
    loads_series['datetime'] = datetime_index

    # Save files
    if scenario_destination_path is not None:
        print('Saving files in zipped csv in "{}"'.format(scenario_destination_path))
        if not os.path.exists(scenario_destination_path):
            os.makedirs(scenario_destination_path)
            
    load_p_forecasted = conso.create_csv(prng, loads_series, scenario_destination_path,
                                        forecasted=True, reordering=True,
                                        shift=True, write_results=write_results, index=False)
    load_p = conso.create_csv(
        prng,
        loads_series, scenario_destination_path,
        reordering=True,
        noise=params['planned_std'],
        write_results=write_results,
        index=False
    )
    
    return load_p, load_p_forecasted



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import json
import os

import numpy as np

import chronix2grid.constants as cst


def parse_seed_arg(seed, arg_name, default_seed):
    if seed is not None:
        try:
            casted_seed = int(seed)
        except TypeError:
            raise RuntimeError(f'The parameter {arg_name} must be an integer')
    else:
        casted_seed = default_seed
    return casted_seed


def generate_default_seed(prng):
    return prng.integers(low=0, high=2 ** 31, dtype=int)


def dump_seeds(output_directory, seeds, scenario_name=''):
    with open(os.path.join(output_directory, scenario_name+'_'+cst.SEEDS_FILE_NAME), 'w') as f:
        json.dump(seeds, f)


# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)


# this generates some chronics for a given environment, provided that all necessary files are present in its repo
import copy
from datetime import datetime, timedelta
import json
import shutil
import time
import pandas as pd
import os
import cvxpy as cp
import grid2op
from grid2op.Parameters import Parameters
from grid2op.Chronics import ChangeNothing, FromNPY
from grid2op.Action import DontAct
from grid2op.Opponent import NeverAttackBudget, BaseOpponent
from lightsim2grid import LightSimBackend
import numpy as np
from numpy.random import default_rng

from chronix2grid.generation.consumption import ConsumptionGeneratorBackend
from chronix2grid.generation.renewable import RenewableBackend
from chronix2grid.generation.dispatch.PypsaDispatchBackend import PypsaDispatcher
from chronix2grid.getting_started.example.input.generation.patterns import ref_pattern_path
from chronix2grid.generation.dispatch.EconomicDispatch import ChroniXScenario

import warnings

FLOATING_POINT_PRECISION_FORMAT = '%.1f'

# TODO allow for a "debug" mode where we can save the values for the prices, the renewables generated, the renewables after dispatch 
# and the renewables after the losses
# TODO add a parameter to generate data more correlated for data in the same area but less correlated within different area.


def get_last_scenario_id(env_chronics_dir):
    """This function return the last scenario id identified.
    
    It only works with scenario id formatted like: `WHATEVER_ScenID`, for example: "2050-01-03_0" or "2050_01_03_0"
    but not "2050-01-03-0"

    It also supposes that the directory containing the chronics exists on the hard drive (even if it's empty)
    
    Parameters
    ----------
    env_chronics_dir : _type_
        _description_
    """
    max_ = None
    list_files = os.listdir(env_chronics_dir)
    for el in list_files:
        this_file_path = os.path.join(env_chronics_dir, el)
        if not os.path.isdir(this_file_path):
            continue
        try:
            *date_, scen_id = el.split("_")
        except ValueError:
            continue
        
        try:
            scen_id = int(scen_id)
        except ValueError:
            continue
        
        if max_ is None:
            max_ = scen_id
            
        if scen_id > max_:
            max_ = scen_id
    if max_ is None:
        max_ = -1
    return max_
            
    
def generate_loads(path_env, load_seed, start_date_dt, end_date_dt, dt, number_of_minutes, generic_params,
                   load_q_from_p_coeff=0.7,
                   day_lag=6):
    """
    This function generates the load for each consumption on a grid

    Parameters
    ----------
    path_env : _type_
        _description_
    load_seed : _type_
        _description_
    start_date_dt : _type_
        _description_
    end_date_dt : _type_
        _description_
    dt : _type_
        _description_
    number_of_minutes : _type_
        _description_
    generic_params : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    with open(os.path.join(path_env, "params_load.json"), "r") as f:
        load_params = json.load(f)
    load_params["start_date"] = start_date_dt
    load_params["end_date"] = end_date_dt
    load_params["dt"] = int(dt)
    load_params["T"] = number_of_minutes
    load_params["planned_std"] = float(generic_params["planned_std"])
    
    loads_charac = pd.read_csv(os.path.join(path_env, "loads_charac.csv"), sep=",")
    load_weekly_pattern = pd.read_csv(os.path.join(ref_pattern_path, "load_weekly_pattern.csv"), sep=",")
    
    load_generator = ConsumptionGeneratorBackend(out_path=None,
                                                 seed=load_seed, 
                                                 params=load_params,
                                                 loads_charac=loads_charac,
                                                 write_results=False,
                                                 load_config_manager=None,
                                                 day_lag=day_lag)
    
    load_p, load_p_forecasted = load_generator.run(load_weekly_pattern=load_weekly_pattern)
    load_q = load_p * load_q_from_p_coeff
    load_q_forecasted = load_p_forecasted * load_q_from_p_coeff
    return load_p, load_q, load_p_forecasted, load_q_forecasted


def generate_renewable_energy_sources(path_env, renew_seed, start_date_dt, end_date_dt, dt, number_of_minutes, generic_params, gens_charac):
    """This function generates the amount of power produced by renewable energy sources (res). 
    
    It serves as a maximum value for the economic dispatch. 

    Parameters
    ----------
    path_env : _type_
        _description_
    renew_seed : _type_
        _description_
    start_date_dt : _type_
        _description_
    end_date_dt : _type_
        _description_
    dt : _type_
        _description_
    number_of_minutes : _type_
        _description_
    generic_params : _type_
        _description_
    gens_charac : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    with open(os.path.join(path_env, "params_res.json"), "r") as f:
        renew_params = json.load(f)
    renew_params["start_date"] = start_date_dt
    renew_params["end_date"] = end_date_dt
    renew_params["dt"] = int(dt)
    renew_params["T"] = number_of_minutes
    renew_params["planned_std"] = float(generic_params["planned_std"])
    solar_pattern = np.load(os.path.join(ref_pattern_path, "solar_pattern.npy"))
    renew_backend = RenewableBackend(out_path=None,
                                     seed=renew_seed,
                                     params=renew_params,
                                     loads_charac=gens_charac,
                                     res_config_manager=None,
                                     write_results=False)
    prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted = renew_backend.run(solar_pattern=solar_pattern)
    return prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted


def generate_economic_dispatch(path_env, start_date_dt, end_date_dt, dt, number_of_minutes, generic_params, 
                               load_p, prod_solar, prod_wind, name_gen, gen_type, scenario_id, final_gen_p, gens_charac):
    """This function emulates a perfect market where all productions need to meet the demand at the minimal cost.
    
    It does not consider limit on powerline, nor contigencies etc. The power network does not exist here. Only the ramps and
    pmin / pmax are important.

    Parameters
    ----------
    path_env : _type_
        _description_
    start_date_dt : _type_
        _description_
    end_date_dt : _type_
        _description_
    dt : _type_
        _description_
    number_of_minutes : _type_
        _description_
    generic_params : _type_
        _description_
    load_p : _type_
        _description_
    prod_solar : _type_
        _description_
    prod_wind : _type_
        _description_
    env : _type_
        _description_
    scenario_id : _type_
        _description_
    final_gen_p : _type_
        _description_
    gens_charac : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    with open(os.path.join(path_env, "params_opf.json"), "r") as f:
        opf_params = json.load(f)
    opf_params["start_date"] = start_date_dt
    opf_params["end_date"] = end_date_dt
    opf_params["dt"] = int(dt)
    opf_params["T"] = number_of_minutes
    opf_params["planned_std"] = float(generic_params["planned_std"])
    
    load = pd.DataFrame(load_p.sum(axis=1))
    total_solar = prod_solar.sum(axis=1)
    total_wind = prod_wind.sum(axis=1)
    
    # init the dispatcher
    gens_charac_this = copy.deepcopy(gens_charac)
    gens_charac_this["pmax"] = gens_charac_this["Pmax"]
    gens_charac_this["pmin"] = gens_charac_this["Pmin"]
    gens_charac_this["cost_per_mw"] = gens_charac_this["marginal_cost"]
    economic_dispatch = PypsaDispatcher.from_dataframe(gens_charac_this)
    
    # need to hack it to work...
    n_gen = len(name_gen)
    gen_p_orig = np.zeros((prod_solar.shape[0], n_gen))
    economic_dispatch._chronix_scenario = ChroniXScenario(loads=1.0 * load_p,
                                                          prods=pd.DataFrame(1.0 * gen_p_orig, columns=name_gen),
                                                          scenario_name=scenario_id,
                                                          res_names={"wind": name_gen[gen_type == "wind"],
                                                                     "solar": name_gen[gen_type == "solar"]
                                                                    }
                                                         )
    economic_dispatch.read_hydro_guide_curves(os.path.join(ref_pattern_path, 'hydro_french.csv'))
    hydro_constraints = economic_dispatch.make_hydro_constraints_from_res_load_scenario()
    res_dispatch = economic_dispatch.run(load * (1.0 + 0.01 * float(opf_params["losses_pct"])),
                                         total_solar,
                                         total_wind,
                                         opf_params,
                                         gen_constraints=hydro_constraints,
                                         pyomo=False,
                                         solver_name="cbc")
    
    if res_dispatch is None:     
        error_ = RuntimeError("Pypsa failed to find a solution")
        return None, None, None, error_
    
    # now assign the results
    final_gen_p = 1.0 * final_gen_p  # copy the data frame to avoid modify the original one
    for gen_id, gen_nm in enumerate(name_gen):
        if gen_nm in res_dispatch.chronix.prods_dispatch:
            final_gen_p.iloc[:, gen_id] = 1.0 * res_dispatch.chronix.prods_dispatch[gen_nm].values
    
    #handle curtailment
    mask_wind = total_wind.values > 0.001
    wind_curt = (res_dispatch.chronix.prods_dispatch['agg_wind'].values[mask_wind] / total_wind.values[mask_wind]).reshape(-1,1)
    final_gen_p.iloc[mask_wind, gen_type == "wind"] *= wind_curt
    
    mask_solar = total_solar.values > 0.001  # be carefull not to divide by 0 in case of solar !
    solar_curt = (res_dispatch.chronix.prods_dispatch['agg_solar'].values[mask_solar] / total_solar.values[mask_solar]).reshape(-1,1)
    final_gen_p.iloc[mask_solar, gen_type == "solar"] *= solar_curt
    
    total_wind_curt = total_wind.values.sum() - res_dispatch.chronix.prods_dispatch['agg_wind'].values.sum()
    total_solar_curt = total_solar.values[mask_solar].sum() - res_dispatch.chronix.prods_dispatch['agg_solar'].values[mask_solar].sum()
    return final_gen_p, total_wind_curt, total_solar_curt, None


def _adjust_gens_old(all_loss_orig,
                 env_for_loss,
                 datetimes,
                 total_solar,
                 total_wind,
                 params,
                 env_path,
                 env_param,
                 load_without_loss,
                 load_p, 
                 load_q,
                 gen_p,
                 gen_v,
                 economic_dispatch,
                 diff_,
                 threshold_stop=0.1,  # stop when all generators move less that this
                 max_iter=100,  # declare a failure after this number of iteration
                 iter_quality_decrease=50,  # acept a reduction of the quality after this number of iteration
                 percentile_quality_decrease=99,
                 ):
    """This function is an auxilliary function.
    
    Like its main one (see handle_losses) it is here to make sure that if you run an AC model with the data generated, 
    then the generator setpoints will not change too much 
    (less than `threshold_stop` MW)

    Parameters
    ----------
    all_loss_orig : _type_
        _description_
    env_for_loss : _type_
        _description_
    datetimes : _type_
        _description_
    total_solar : _type_
        _description_
    total_wind : _type_
        _description_
    params : _type_
        _description_
    env_path : _type_
        _description_
    env_param : _type_
        _description_
    load_without_loss : _type_
        _description_
    load_p : _type_
        _description_
    load_q : _type_
        _description_
    gen_p : _type_
        _description_
    gen_v : _type_
        _description_
    economic_dispatch : _type_
        _description_
    diff_ : _type_
        _description_
    threshold_stop : float, optional
        _description_, by default 0.1

    Returns
    -------
    _type_
        _description_
    """
    quality_ = None
    error_ = None
    if np.any(~np.isfinite(gen_p)):
        error_ = RuntimeError("Input data contained Nans !")
        return None, error_, quality_
    all_loss = all_loss_orig
    res_gen_p = 1.0 * gen_p
    iter_num = 0
    hydro_constraints = economic_dispatch.make_hydro_constraints_from_res_load_scenario()
    while True:
        iter_num += 1
        load = load_without_loss + all_loss
        load = pd.DataFrame(load.ravel(), index=datetimes)
        
        # "never" decrease (during iteration) some generators
        min__ = diff_.min()  # this is negative
        gen_max_pu_t = None
        gen_min_pu_t = {gen_nm: np.maximum((res_gen_p[:,gen_id] + min__) / economic_dispatch.generators.loc[gen_nm].p_nom,
                                            env_for_loss.gen_pmin[gen_id] / economic_dispatch.generators.loc[gen_nm].p_nom
                                            )
                        for gen_id, gen_nm in enumerate(env_for_loss.name_gen) if env_for_loss.gen_redispatchable[gen_id]}
        
        ### run the dispatch with the loss
        dispatch_res = economic_dispatch.run(load,
                                             total_solar=total_solar,
                                             total_wind=total_wind,
                                             params=params,
                                             pyomo=False,
                                             solver_name="cbc",
                                             gen_constraints=copy.deepcopy(hydro_constraints),
                                             gen_max_pu_t=gen_max_pu_t,
                                             gen_min_pu_t=gen_min_pu_t,
                                             )
        
        if dispatch_res is None:     
            error_ = RuntimeError(f"Pypsa failed to find a solution at iteration {iter_num}")
            break
        
        # assign the generators
        for gen_id, gen_nm in enumerate(env_for_loss.name_gen):
            if gen_nm in dispatch_res.chronix.prods_dispatch:
                res_gen_p[:, gen_id] = 1.0 * dispatch_res.chronix.prods_dispatch[gen_nm].values
                  
        #handle wind curtailment
        mask_winds = total_wind.values > 0.001
        res_gen_p[mask_winds, :][:,env_for_loss.gen_type == "wind"] *= (dispatch_res.chronix.prods_dispatch['agg_wind'].values[mask_winds] / total_wind.values[mask_winds]).reshape(-1,1)
        total_wind.loc[mask_winds] = 1.0 * dispatch_res.chronix.prods_dispatch["agg_wind"].values[mask_winds]
        
        # re evaluate the losses
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            env_fixed = grid2op.make(
                env_path,
                test=True,
                # grid_path=grid_path, # assign it the 118 grid
                param=env_param,
                backend=LightSimBackend(),
                chronics_class=FromNPY,
                # chronics_path=path_chronix2grid,
                data_feeding_kwargs={"load_p": load_p,
                                     "load_q": load_q,
                                     "prod_p": 1.0 * res_gen_p,
                                     "prod_v": gen_v}
                )
        diff_ = np.full((env_fixed.max_episode_duration(), env_fixed.n_gen), fill_value=np.NaN)
        all_loss[:] = np.NaN
        
        i = 0
        obs = env_fixed.reset()
        all_loss[i] = np.sum(obs.gen_p) - np.sum(obs.load_p)
        diff_[i] = obs.gen_p - res_gen_p[i]

        
        done = False
        while not done:
            obs, reward, done, info = env_fixed.step(env_fixed.action_space())
            i += 1
            if done:
                break
            all_loss[i] = np.sum(obs.gen_p) - np.sum(obs.load_p)
            diff_[i] = obs.gen_p - res_gen_p[i]
        
        max_diff_ = np.abs(diff_).max()
        if not np.isfinite(max_diff_):
            error_ = RuntimeError(f"Some nans were found in the generated data at iteration {iter_num}")
            res_gen_p = None
            quality_ = None
            break
            
            
        if max_diff_ <= threshold_stop:
            quality_ = (iter_num,
                        float(np.mean(np.abs(diff_))),
                        float(np.percentile(np.abs(diff_), 95)),
                        float(np.percentile(np.abs(diff_), 99)),
                        float(max_diff_)
            )
            break
        
        if iter_num >= iter_quality_decrease:
            quantile = np.percentile(np.abs(diff_), percentile_quality_decrease)
            if quantile <= threshold_stop:
                quality_ = (iter_num,
                            float(np.mean(np.abs(diff_))),
                            float(np.percentile(np.abs(diff_), 95)),
                            float(np.percentile(np.abs(diff_), 99)),
                            float(np.max(np.abs(diff_)))
                )
                break
                    
        if iter_num >= max_iter:
            error_ = RuntimeError("Too much iterations performed when adjusting for the losses")
            res_gen_p = None
            quality_ = None
            break
        
    return res_gen_p, error_, quality_


def _adjust_gens(all_loss_orig,
                 env_for_loss,
                 datetimes,
                 total_solar,
                 total_wind,
                 params,
                 env_path,
                 env_param,
                 load_without_loss,
                 load_p, 
                 load_q,
                 gen_p,
                 gen_v,
                 economic_dispatch,
                 diff_,
                 threshold_stop=0.1,  # stop when all generators move less that this
                 max_iter=100,  # declare a failure after this number of iteration
                 iter_quality_decrease=50,  # acept a reduction of the quality after this number of iteration
                 percentile_quality_decrease=99,
                 ):
    """This function is an auxilliary function.
    
    Like its main one (see handle_losses) it is here to make sure that if you run an AC model with the data generated, 
    then the generator setpoints will not change too much 
    (less than `threshold_stop` MW)

    Parameters
    ----------
    all_loss_orig : _type_
        _description_
    env_for_loss : _type_
        _description_
    datetimes : _type_
        _description_
    total_solar : _type_
        _description_
    total_wind : _type_
        _description_
    params : _type_
        _description_
    env_path : _type_
        _description_
    env_param : _type_
        _description_
    load_without_loss : _type_
        _description_
    load_p : _type_
        _description_
    load_q : _type_
        _description_
    gen_p : _type_
        _description_
    gen_v : _type_
        _description_
    economic_dispatch : _type_
        _description_
    diff_ : _type_
        _description_
    threshold_stop : float, optional
        _description_, by default 0.1

    Returns
    -------
    _type_
        _description_
    """
    quality_ = None
    error_ = None
    if np.any(~np.isfinite(gen_p)):
        error_ = RuntimeError("Input data contained Nans !")
        return None, error_, quality_
    all_loss = all_loss_orig
    res_gen_p = 1.0 * gen_p
    iter_num = 0
    hydro_constraints = economic_dispatch.make_hydro_constraints_from_res_load_scenario()
    
    # defined some global variable (used for all optimization problems)
    turned_off_orig = 1.0 * (gen_p[:, env_for_loss.gen_redispatchable] == 0.)
    ids_hyrdo = []
    total_gen = np.sum(env_for_loss.gen_redispatchable)
    total_step = total_solar.shape[0]
    gen_id = 0
    for i in range(env_for_loss.n_gen):
        if env_for_loss.gen_redispatchable[i]:
            if env_for_loss.gen_type[i] == "hydro":
                ids_hyrdo.append(gen_id)
            gen_id += 1
    ids_hyrdo = np.array(ids_hyrdo)
    
    # define the constraints    
    scaling_factor = env_for_loss.gen_pmax[env_for_loss.gen_redispatchable]
    p_min = np.repeat(env_for_loss.gen_pmin[env_for_loss.gen_redispatchable].reshape(1,-1)  / scaling_factor,
                        total_step,
                        axis=0)
    p_max = np.repeat(env_for_loss.gen_pmax[env_for_loss.gen_redispatchable].reshape(1,-1) * params["PmaxErrorCorrRatio"] / scaling_factor,
                        total_step,
                        axis=0)
    
    ramp_min = np.repeat(-env_for_loss.gen_max_ramp_down[env_for_loss.gen_redispatchable].reshape(1,-1) * params["RampErrorCorrRatio"]  / scaling_factor,
                            total_step - 1,
                            axis=0)
    ramp_max = np.repeat(env_for_loss.gen_max_ramp_up[env_for_loss.gen_redispatchable].reshape(1,-1) * params["RampErrorCorrRatio"]  / scaling_factor,
                            total_step - 1,
                            axis=0)
    p_max[:, ids_hyrdo] = 1.0 * hydro_constraints["p_max_pu"].values
     
    while True:
        iter_num += 1
        load = load_without_loss + all_loss
        load = pd.DataFrame(load.ravel(), index=datetimes)
        
        # "never" decrease (during iteration) some generators
        min__ = diff_.min()  # this is negative
        load = load_without_loss + all_loss - np.sum(res_gen_p[:,~env_for_loss.gen_redispatchable], axis=1)
        scale_for_loads =  np.repeat(scaling_factor.reshape(1,-1), total_step, axis=0)
        target_vector = res_gen_p[:,env_for_loss.gen_redispatchable] / scaling_factor         
        
        #### cvxpy
        p_t = cp.Variable(shape=(total_step,total_gen), pos=True)
        real_p = cp.multiply(p_t, scale_for_loads)
        
        constraints = [p_t >= p_min,
                       p_t <= p_max,
                       p_t[1:,:] - p_t[:-1,:] >= ramp_min,
                       p_t[1:,:] - p_t[:-1,:] <= ramp_max,
                       cp.sum(real_p, axis=1) == load.reshape(-1),
                      ]
        cost = cp.sum_squares(p_t - target_vector) + cp.norm1(cp.multiply(p_t, turned_off_orig))
        prob = cp.Problem(cp.Minimize(cost), constraints)
        try:
            prob.solve()
        except cp.error.SolverError as exc_:
            error_ = RuntimeError(f"cvxpy failed to find a solution at iteration {iter_num}, error {exc_}")
            res_gen_p = None
            quality_ = None
            break
            
        # assign the generators
        gen_p_after_optim = real_p.value
        if gen_p_after_optim is None:
            error_ = RuntimeError(f"cvxpy failed to find a solution at iteration {iter_num}, and returned None.")
            res_gen_p = None
            quality_ = None
            break
            
        id_redisp = 0
        for gen_id, gen_nm in enumerate(env_for_loss.name_gen):
            if env_for_loss.gen_redispatchable[gen_id]:
                res_gen_p[:, gen_id] = 1.0 * gen_p_after_optim[:, id_redisp]
                id_redisp += 1
        
        # re evaluate the losses
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            env_fixed = grid2op.make(
                env_path,
                test=True,
                # grid_path=grid_path, # assign it the 118 grid
                param=env_param,
                backend=LightSimBackend(),
                chronics_class=FromNPY,
                # chronics_path=path_chronix2grid,
                data_feeding_kwargs={"load_p": load_p,
                                     "load_q": load_q,
                                     "prod_p": 1.0 * res_gen_p,
                                     "prod_v": gen_v},
                opponent_budget_per_ts=0.,
                opponent_init_budget=0.,
                opponent_class=BaseOpponent,
                opponent_budget_class=NeverAttackBudget,
                opponent_action_class=DontAct,
                )
        diff_ = np.full((env_fixed.max_episode_duration(), env_fixed.n_gen), fill_value=np.NaN)
        all_loss[:] = np.NaN
        
        i = 0
        obs = env_fixed.reset()
        all_loss[i] = np.sum(obs.gen_p) - np.sum(obs.load_p)
        diff_[i] = obs.gen_p - res_gen_p[i]

        done = False
        while not done:
            obs, reward, done, info = env_fixed.step(env_fixed.action_space())
            i += 1
            if done:
                break
            all_loss[i] = np.sum(obs.gen_p) - np.sum(obs.load_p)
            diff_[i] = obs.gen_p - res_gen_p[i]
        
        max_diff_ = np.abs(diff_).max()
        print(f"{iter_num = } : {max_diff_ = :.2f}")
        if not np.isfinite(max_diff_):
            error_ = RuntimeError(f"Some nans were found in the generated data at iteration {iter_num}")
            res_gen_p = None
            quality_ = None
            break
            
        if max_diff_ <= threshold_stop:
            quality_ = (iter_num,
                        float(np.mean(np.abs(diff_))),
                        float(np.percentile(np.abs(diff_), 95)),
                        float(np.percentile(np.abs(diff_), 99)),
                        float(max_diff_)
            )
            break
        
        if iter_num >= iter_quality_decrease:
            quantile = np.percentile(np.abs(diff_), percentile_quality_decrease)
            if quantile <= threshold_stop:
                quality_ = (iter_num,
                            float(np.mean(np.abs(diff_))),
                            float(np.percentile(np.abs(diff_), 95)),
                            float(np.percentile(np.abs(diff_), 99)),
                            float(np.max(np.abs(diff_)))
                )
                break
                    
        if iter_num >= max_iter:
            error_ = RuntimeError("Too much iterations performed when adjusting for the losses")
            res_gen_p = None
            quality_ = None
            break
        
    return res_gen_p, error_, quality_


def _fix_losses_one_scenario(env_for_loss,
                            scenario_id,
                            params,
                            env_path,
                            env_param,
                            load_df,
                            threshold_stop=0.05,  # decide I stop when the data move of less of 0.5 MW at maximum
                            max_iter=100,  # maximum number of iteration
                            iter_quality_decrease=20,  # after 20 iteration accept a degradation in the quality
                            percentile_quality_decrease=99,  # replace the "at maximum" by "percentile 99%"
                            ):
    """This function is an auxilliary function.
    
    Like its main one (see handle_losses) it is here to make sure that if you run an AC model with the data generated, 
    then the generator setpoints will not change too much 
    (less than `threshold_stop` MW)

    Parameters
    ----------
    env_for_loss : _type_
        _description_
    scenario_id : _type_
        _description_
    params : _type_
        _description_
    env_path : _type_
        _description_
    env_param : _type_
        _description_
    load_df : _type_
        _description_
    threshold_stop : float, optional
        _description_, by default 0.5
    max_iter : int, optional
        _description_, by default 100

    Returns
    -------
    _type_
        _description_
    """
    gen_p_orig = np.full((env_for_loss.max_episode_duration(), env_for_loss.n_gen), fill_value=np.NaN, dtype=np.float32)
    final_gen_v = np.full((env_for_loss.max_episode_duration(), env_for_loss.n_gen), fill_value=np.NaN, dtype=np.float32)
    final_load_p = np.full((env_for_loss.max_episode_duration(), env_for_loss.n_load), fill_value=np.NaN, dtype=np.float32)
    final_load_q = np.full((env_for_loss.max_episode_duration(), env_for_loss.n_load), fill_value=np.NaN, dtype=np.float32)
    all_loss_orig = np.zeros(env_for_loss.max_episode_duration())
    max_diff_orig = np.zeros(env_for_loss.max_episode_duration())
    datetimes = np.zeros(env_for_loss.max_episode_duration(), dtype=datetime)
    
    env_for_loss.set_id(scenario_id)
    obs = env_for_loss.reset()
    
    i = 0
    all_loss_orig[i] = np.sum(obs.gen_p) - np.sum(obs.load_p)
    final_gen_v[i] = obs.gen_v
    final_load_p[i] = obs.load_p
    final_load_q[i] = obs.load_q
    gen_p_orig[i] = 1.0 * obs.gen_p
    datetimes[i] = obs.get_time_stamp()
    max_diff_orig[i] = np.max(np.abs(obs.gen_p -  env_for_loss.chronics_handler.real_data._prod_p[i]))
    
    done = False
    while not done:
        obs, reward, done, info = env_for_loss.step(env_for_loss.action_space())
        if done:
            break
        i += 1
        all_loss_orig[i] = np.sum(obs.gen_p) - np.sum(obs.load_p)
        final_load_p[i] = 1.0 * obs.load_p
        final_load_q[i] = 1.0 * obs.load_p
        gen_p_orig[i] = env_for_loss.chronics_handler.real_data._prod_p[i]  # 1.0 * obs.gen_p
        datetimes[i] = obs.get_time_stamp()
        max_diff_orig[i] = np.max(np.abs(obs.gen_p -  env_for_loss.chronics_handler.real_data._prod_p[i]))
        
    total_solar = np.sum(gen_p_orig[:, env_for_loss.gen_type == "solar"], axis=1)
    total_wind = np.sum(gen_p_orig[:, env_for_loss.gen_type == "wind"], axis=1)
    load_without_loss = np.sum(final_load_p, axis=1) #  - total_solar - total_wind
    
    # load the right data
    df = pd.read_csv(os.path.join(env_path, "prods_charac.csv"), sep=",")
    df["pmax"] = df["Pmax"]
    df["pmin"] = df["Pmin"]
    df["cost_per_mw"] = df["marginal_cost"]
    economic_dispatch = PypsaDispatcher.from_dataframe(df)
    economic_dispatch.read_hydro_guide_curves(os.path.join(ref_pattern_path, 'hydro_french.csv'))
    economic_dispatch._chronix_scenario = ChroniXScenario(loads=1.0 * load_df,
                                                          prods=pd.DataFrame(1.0 * gen_p_orig, columns=env_for_loss.name_gen),
                                                          scenario_name=scenario_id,
                                                          res_names= {"wind": env_for_loss.name_gen[env_for_loss.gen_type == "wind"],
                                                                      "solar": env_for_loss.name_gen[env_for_loss.gen_type == "solar"]
                                                          }
                                                         )
    
    error_ = None
    total_solar_orig = pd.Series(total_solar.ravel(), index=datetimes)
    total_wind_orig = pd.Series(total_wind.ravel(), index=datetimes)
    
    total_solar = 1.0 * total_solar_orig
    total_wind = 1.0 * total_wind_orig
    res_gen_p = 1.0 * gen_p_orig
    diff_ = 1.0 * max_diff_orig
    diff_ = diff_.reshape(-1,1)
    res_gen_p, error_, quality_ = _adjust_gens(all_loss_orig,
                                               env_for_loss,
                                               datetimes,
                                               total_solar,
                                               total_wind,
                                               params,
                                               env_path,
                                               env_param,
                                               load_without_loss,
                                               final_load_p, 
                                               final_load_q,
                                               gen_p_orig,
                                               final_gen_v,
                                               economic_dispatch,
                                               diff_,
                                               threshold_stop=threshold_stop,
                                               max_iter=max_iter,
                                               iter_quality_decrease=iter_quality_decrease,
                                               percentile_quality_decrease=percentile_quality_decrease)
    
    if error_ is not None:
        # the procedure failed
        return None, error_, None
    
    return res_gen_p, error_, quality_


def handle_losses(path_env,
                  n_gen,
                  name_gen,
                  gens_charac,
                  load_p,
                  load_q,
                  final_gen_p,
                  start_date_dt,
                  dt_dt,
                  scenario_id, 
                  PmaxErrorCorrRatio=0.9,
                  RampErrorCorrRatio=0.95,
                  threshold_stop=0.05,
                  max_iter=100,
                  iter_quality_decrease=20,  # after 20 iteration accept a degradation in the quality
                  percentile_quality_decrease=99,  # replace the "at maximum" by "percentile 99%"
                  ):
    """This function is here to make sure that if you run an AC model with the data generated, then the generator setpoints will not change too much 
    (less than `threshold_stop` MW)

    Parameters
    ----------
    path_env : _type_
        _description_
    n_gen : _type_
        _description_
    gens_charac : _type_
        _description_
    load_p : _type_
        _description_
    load_q : _type_
        _description_
    final_gen_p : _type_
        _description_
    start_date_dt : _type_
        _description_
    dt_dt : _type_
        _description_
    scenario_id : _type_
        _description_
    PmaxErrorCorrRatio : float, optional
        _description_, by default 0.9
    RampErrorCorrRatio : float, optional
        _description_, by default 0.95
    threshold_stop : float, optional
        _description_, by default 0.5
    max_iter : int, optional
        _description_, by default 100

    Returns
    -------
    _type_
        _description_
    """
    
    with open(os.path.join(path_env, "params_opf.json"), "r") as f:
        loss_param = json.load(f)
    loss_param["loss_pct"] = 0.  # losses are handled better in this function
    loss_param["PmaxErrorCorrRatio"] = PmaxErrorCorrRatio
    loss_param["RampErrorCorrRatio"] = RampErrorCorrRatio
    
    # do not treat the slack differently
    loss_param["slack_ramp_limit_ratio"] = loss_param["RampErrorCorrRatio"]
    if "slack_pmin" in loss_param:
        del loss_param["slack_pmin"]
    if "slack_pmax" in loss_param:
        del loss_param["slack_pmax"]
        
    env_param = Parameters()
    env_param.NO_OVERFLOW_DISCONNECTION = True
    gen_v = np.tile(np.array([float(gens_charac.loc[gens_charac["name"] == nm_gen].V) for nm_gen in name_gen ]),
                    load_p.shape[0]).reshape(-1, n_gen)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        env_for_loss = grid2op.make(
            path_env,
            test=True,
            # grid_path=grid_path, # assign it the 118 grid
            param=env_param,
            backend=LightSimBackend(),
            chronics_class=FromNPY,
            # chronics_path=path_chronix2grid,
            data_feeding_kwargs={"load_p": load_p.values,  # np.concatenate([load_p.values[0].reshape(1,-1), load_p.values]),
                                 "load_q": load_q.values,  # np.concatenate([load_q.values[0].reshape(1,-1), load_q.values]),
                                 "prod_p": 1.0 * final_gen_p.values,  # 1.0 * np.concatenate([final_gen_p.values[0].reshape(1,-1), final_gen_p.values]),
                                 "prod_v": gen_v,  # np.concatenate([gen_v[0].reshape(1,-1), gen_v])}
                                 "start_datetime": start_date_dt,
                                 "time_interval": dt_dt,
            }
            )
    res_gen_p, error_, quality_ = _fix_losses_one_scenario(env_for_loss,
                                                           scenario_id,
                                                           loss_param,
                                                           path_env,
                                                           env_for_loss.parameters,
                                                           load_df=load_p,
                                                           threshold_stop=threshold_stop,
                                                           max_iter=max_iter,
                                                           # after 20 iteration accept a degradation in the quality
                                                           iter_quality_decrease=iter_quality_decrease,  
                                                           # replace the "at maximum" by "percentile 99%"
                                                           percentile_quality_decrease=percentile_quality_decrease,  
                                                           )
    if error_ is not None:
        return None, error_, None
    
    # reformat the generators
    res_gen_p_df = pd.DataFrame(res_gen_p, index=final_gen_p.index, columns=env_for_loss.name_gen)
    
    return res_gen_p_df, error_, quality_

def save_generated_data(this_scen_path,
                        load_p,
                        load_p_forecasted,
                        load_q,
                        load_q_forecasted,
                        prod_p_generated,
                        prod_p_after_dispatch,
                        prod_p,  # after everything, the one the agent sees
                        prod_p_forecasted,
                        debug,
                        sep=';',
                        float_prec=FLOATING_POINT_PRECISION_FORMAT):
    """This function saves the data that have been generated by this script.

    Parameters
    ----------
    this_scen_path : _type_
        _description_
    load_p : _type_
        _description_
    load_p_forecasted : _type_
        _description_
    load_q : _type_
        _description_
    load_q_forecasted : _type_
        _description_
    prod_p_generated: _type_
        generators just after renewable generation (non renewables gens are Nans here !)
    prod_p : _type_
        generators after dispatch and losses !
    prod_p_forecasted : _type_
        _description_
    sep : str, optional
        _description_, by default ';'
    float_prec : _type_, optional
        _description_, by default FLOATING_POINT_PRECISION_FORMAT
    """
    li_dfs = [load_p, load_p_forecasted, load_q, load_q_forecasted, prod_p, prod_p_forecasted]
    li_nms = ["load_p", "load_p_forecasted", "load_q", "load_q_forecasted", "prod_p", "prod_p_forecasted"]
    if debug:
        li_dfs.append(prod_p_generated)
        li_nms.append("prod_p_generated")
        li_dfs.append(prod_p_after_dispatch)
        li_nms.append("prod_p_after_dispatch")
    for df, nm in zip(li_dfs, li_nms):
        df.to_csv(os.path.join(this_scen_path, f'{nm}.csv.bz2'),
                  sep=sep,
                  float_format=float_prec,
                  header=True,
                  index=False)       


def save_meta_data(this_scen_path,
                   path_env,
                   start_date_dt,
                   dt_dt : timedelta,
                   load_seed,
                   renew_seed,
                   gen_p_forecast_seed,
                   quality,
                   total_load,
                   total_gen,
                   losses_mwh,
                   losses_avg,
                   wind_curtailed_opf,
                   wind_curtailed_losses,
                   solar_curtailed_opf,
                   generation_time,
                   saving_time,
                   files_to_copy=("maintenance_meta.json",)):
    """This function saves the "meta data" required for a succesful grid2op run !

    Parameters
    ----------
    this_scen_path : _type_
        _description_
    path_env : _type_
        _description_
    start_date_dt : _type_
        _description_
    dt_dt : timedelta
        _description_
    load_seed : _type_
        _description_
    renew_seed : _type_
        _description_
    gen_p_forecast_seed : _type_
        _description_
    quality: tuple

    """
    with open(os.path.join(this_scen_path, "time_interval.info"), "w", encoding="utf-8") as f:
        # f.write(datetime.strftime(dt_dt, "%H:%M"))  # what I want to do but cannot (TypeError: descriptor 'strftime' for 'datetime.date' objects doesn't apply to a 'datetime.timedelta' object)
        total_s = dt_dt.total_seconds()
        hours = total_s // 3600
        mins = (total_s - 3600 * hours) // 60
        f.write(f"{int(hours):02d}:{int(mins):02d}")
        
    with open(os.path.join(this_scen_path, "start_datetime.info"), "w", encoding="utf-8") as f:
        f.write(datetime.strftime(start_date_dt, "%Y-%m-%d %H:%M"))
    with open(os.path.join(this_scen_path, "_seeds_info.json"), "w", encoding="utf-8") as f:
        json.dump(obj={"load_seed": int(load_seed), "renew_seed": int(renew_seed), "gen_p_forecast_seed": int(gen_p_forecast_seed)},
                  fp=f)
    with open(os.path.join(this_scen_path, "generation_quality.json"), "w", encoding="utf-8") as f:
        iter_num, mean_, percent_95, percent_99, max_ = quality
        json.dump(obj={"iter_num": int(iter_num),
                       "avg": float(mean_),
                       "percent_95": float(percent_95),
                       "percent_99": float(percent_99),
                       "max": float(max_), 
                       "info": ("avg, percent_95, percent_99, max: this 'quality' is the difference between the DC solver and the AC solver. This is the number of "
                                "MW that will differ from the grid2op observation compared to the generated data by chronix2grid.",
                                "total_load, total_gen: total amount of energy consumed / produced for the generated scenario.",
                                "losses_mwh: total amount of losses for the scenario (in energy)",
                                "losses_avg: average (per step) of the loss (avg[loss_this_step / total_generation_this_step])",
                                "wind_curtailed_opf: total (in energy) wind power curtailed by the OPF",
                                "wind_curtailed_losses: total (in energy) wind power curtailed by the loss",
                                "solar_curtailed_opf: total (in energy) solar power curtailed by the OPF",
                                "iter_num: number of iteration of the loss algorithm",
                                "generation_time: total time spent to generate these data (in seconds)",
                                "saving_time: total time spent to save the generated data (in seconds), this excludes the metadata saving time",
                                ),
                       "total_load": float(total_load),
                       "total_gen": float(total_gen),
                       "losses_mwh": float(losses_mwh),
                       "losses_avg": float(losses_avg),
                       "wind_curtailed_opf": float(wind_curtailed_opf),
                       "wind_curtailed_losses": float(wind_curtailed_losses),
                       "solar_curtailed_opf": float(solar_curtailed_opf),
                       "generation_time": float(generation_time),
                       "saving_time": float(saving_time),
                       },
                  fp=f,
                  sort_keys=True,
                  indent=4)
    
    for fn_ in files_to_copy:
        src_path = os.path.join(path_env, fn_)
        if os.path.exists(src_path):
            shutil.copy(src=src_path, 
                        dst=os.path.join(this_scen_path, fn_))
    

def generate_a_scenario(path_env,
                        name_gen,
                        gen_type,
                        output_dir, 
                        start_date,
                        dt,
                        scen_id,
                        load_seed,
                        renew_seed,
                        gen_p_forecast_seed,
                        handle_loss=True,
                        nb_steps=None,
                        PmaxErrorCorrRatio=0.9,
                        RampErrorCorrRatio=0.95,
                        threshold_stop=0.05,
                        max_iter=100,
                        debug=True  # TODO more feature !
                        ):
    """This function generates and save the data for a scenario.
    
    Generation includes:
    - load active value
    - load reactive value
    - renewable generation
    - controlable generation

    Put "outputdir=None" if you don't want to save the data.
    
    
    Parameters
    ----------
    env_name : _type_
        _description_
    name_gen : _type_
        _description_
    gen_type: _type_
        _description_
    output_dir : _type_
        _description_
    start_date : _type_
        _description_
    dt : _type_
        _description_
    scen_id : _type_
        _description_
    load_seed : _type_
        _description_
    renew_seed : _type_
        _description_
    gen_p_forecast_seed : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    beg_ = time.perf_counter()
    scenario_id = f"{start_date}_{scen_id}"
    dt_dt = timedelta(minutes=int(dt))
    start_date_dt = datetime.strptime(start_date, "%Y-%m-%d") - dt_dt
    if nb_steps is None:
        end_date_dt = start_date_dt + timedelta(days=7) + 2 * dt_dt
    else:
        end_date_dt = start_date_dt + (int(nb_steps) + 2) * dt_dt
    end_date = datetime.strftime(end_date_dt,  "%Y-%m-%d %H:%M:%S")
    with open(os.path.join(path_env, "params.json"), "r") as f:
        generic_params = json.load(f)
    number_of_minutes = int((end_date_dt - start_date_dt).total_seconds() // 60)
    gens_charac = pd.read_csv(os.path.join(path_env, "prods_charac.csv"), sep=",")
    
    # conso generation
    load_p, load_q, load_p_forecasted, load_q_forecasted = generate_loads(path_env,
                                                                          load_seed,
                                                                          start_date_dt,
                                                                          end_date_dt,
                                                                          dt,
                                                                          number_of_minutes,
                                                                          generic_params,
                                                                          day_lag=6  # TODO 6 because it's 2050
                                                                          )
    
    # renewable energy sources generation
    res_renew = generate_renewable_energy_sources(path_env,renew_seed, start_date_dt, end_date_dt, dt, number_of_minutes, generic_params, gens_charac)
    prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted = res_renew

    if prod_solar.isna().any().any():
        error_ = RuntimeError("Nan generated in solar data")
        return error_, None, None, None, None, None, None, None
    if prod_wind.isna().any().any():
        error_ = RuntimeError("Nan generated in wind data")
        return error_, None, None, None, None, None, None, None
    
    # create the result data frame for the generators
    final_gen_p = pd.merge(prod_solar, prod_wind, left_index=True, right_index=True)
    
    for el in name_gen:
        if el in final_gen_p:
            continue
        final_gen_p[str(el)] = np.NaN
    final_gen_p = final_gen_p[name_gen]
    
    # generate economic dispatch
    res_disp = generate_economic_dispatch(path_env, start_date_dt, end_date_dt, dt, number_of_minutes, generic_params,
                                          load_p, prod_solar, prod_wind, name_gen, gen_type, scenario_id, final_gen_p, gens_charac)
    gen_p_after_dispatch, total_wind_curt_opf, total_solar_curt_opf, error_ = res_disp
    
    if error_ is not None:
        # TODO log that !
        return error_, None, None, None, None, None, None, None
    
    # now try to move the generators so that when I run an AC powerflow, the setpoint of generators does not change "too much"
    if handle_loss:
        n_gen = len(name_gen)
        res_gen_p_df, error_, quality_ = handle_losses(path_env,
                                                       n_gen,
                                                       name_gen,
                                                       gens_charac,
                                                       load_p,
                                                       load_q,
                                                       gen_p_after_dispatch,
                                                       start_date_dt,
                                                       dt_dt,
                                                       scenario_id, 
                                                       PmaxErrorCorrRatio=PmaxErrorCorrRatio,
                                                       RampErrorCorrRatio=RampErrorCorrRatio,
                                                       threshold_stop=threshold_stop,
                                                       max_iter=max_iter)
        if error_ is not None:
            # TODO log that !
            return error_, None, None, None, None, None, None, None
    else:
        res_gen_p_df = 1.0 * gen_p_after_dispatch
        quality_ = (-1, float("Nan"), float("Nan"), float("Nan"), float("Nan"))
    
    prng = default_rng(gen_p_forecast_seed)
    res_gen_p_forecasted_df = res_gen_p_df * prng.lognormal(mean=0.0, sigma=float(generic_params["planned_std"]), size=res_gen_p_df.shape)
    res_gen_p_forecasted_df = res_gen_p_forecasted_df.shift(-1)
    res_gen_p_forecasted_df.iloc[-1] = 1.0 * res_gen_p_forecasted_df.iloc[-2]
    end_ = time.perf_counter()
    if output_dir is not None:
        beg_save = time.perf_counter()
        this_scen_path = os.path.join(output_dir, scenario_id)
        if not os.path.exists(this_scen_path):
            os.mkdir(this_scen_path)
        save_generated_data(this_scen_path,
                            load_p,
                            load_p_forecasted,
                            load_q,
                            load_q_forecasted,
                            final_gen_p,  # generated, before economic dispatch
                            gen_p_after_dispatch,  # generated, after economic dispatch (and possibly curtailment)
                            res_gen_p_df,
                            res_gen_p_forecasted_df,
                            debug=debug)
        total_load = float(load_p.sum().sum())
        total_gen = float(res_gen_p_df.sum().sum())
        gen_p_per_step = res_gen_p_df.sum(axis=1)
        proper_MWh_unit = float(dt_dt.total_seconds() / 3600.)
        wind_curtailed_losses = (gen_p_after_dispatch.iloc[:, gen_type=="wind"].sum().sum() - res_gen_p_df.iloc[:, gen_type=="wind"].sum().sum())
        end_save = time.perf_counter()
        save_meta_data(this_scen_path,
                       path_env,
                       start_date_dt,
                       dt_dt,
                       load_seed,
                       renew_seed,
                       gen_p_forecast_seed,
                       quality_,
                       total_load=total_load * proper_MWh_unit,
                       total_gen=total_gen * proper_MWh_unit,
                       losses_mwh=(total_gen - total_load) * proper_MWh_unit,
                       losses_avg=np.mean((gen_p_per_step - load_p.sum(axis=1)) / gen_p_per_step),
                       wind_curtailed_opf=float(total_wind_curt_opf * proper_MWh_unit),
                       wind_curtailed_losses=float(wind_curtailed_losses * proper_MWh_unit),
                       solar_curtailed_opf=float(total_solar_curt_opf * proper_MWh_unit),
                       generation_time=end_ - beg_,
                       saving_time=end_save - beg_save
                       )
        
    return error_, quality_, load_p, load_p_forecasted, load_q, load_q_forecasted, res_gen_p_df, res_gen_p_forecasted_df



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import calendar
from datetime import datetime
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from .. import generation_utils as utils
import chronix2grid.constants as cst

def compute_loads(loads_charac, temperature_noise, params, load_weekly_pattern, start_day, add_dim, day_lag=0):
    #6  # this is only TRUE if you simulate 2050 !!! formula does not really work
    
    # Compute active part of loads
    weekly_pattern = load_weekly_pattern['test'].values
    
    # start_day_of_week = start_day.weekday()
    # first_dow_chronics = datetime.strptime(load_weekly_pattern["datetime"].iloc[1], "%Y-%m-%d %H:%M:%S").weekday()
    # + (calendar.isleap(start_day.year) if start_day.month >= 3 else 0)
    # day_lag = (first_dow_chronics - start_day_of_week) % 7
    # day_lag = 0
    loads_series = {}
    for i, name in enumerate(loads_charac['name']):
        mask = (loads_charac['name'] == name)
        if loads_charac[mask]['type'].values == 'residential':
            locations = [loads_charac[mask]['x'].values[0], loads_charac[mask]['y'].values[0]]
            Pmax = loads_charac[mask]['Pmax'].values[0]
            loads_series[name] = compute_residential(locations, Pmax, temperature_noise, params, weekly_pattern, index=i, day_lag=day_lag, add_dim=add_dim)

        if loads_charac[mask]['type'].values == 'industrial':
            raise NotImplementedError("Impossible to generate industrial loads for now.")
            Pmax = loads_charac[mask]['Pmax'].values[0]
            loads_series[name] = compute_industrial(Pmax, params)
    return loads_series

def compute_residential(locations, Pmax, temperature_noise, params, weekly_pattern, index, day_lag=None, add_dim=0):


    # Compute refined signals
    temperature_signal = utils.interpolate_noise(
        temperature_noise,
        params,
        locations,
        time_scale=params['temperature_corr'],
        add_dim=add_dim)
    temperature_signal = temperature_signal.astype(float)
    
    # Compute seasonal pattern
    Nt_inter = int(params['T'] // params['dt'] + 1)
    
    # t = np.linspace(0, params['T'], Nt_inter, endpoint=True)
    t = np.linspace(0., (params['end_date'] - params["start_date"]).total_seconds(), Nt_inter, endpoint=True, dtype=float)
    
    start_year = pd.to_datetime(str(params['start_date'].year) + '/01/01', format='%Y-%m-%d')
    start_min = float(pd.Timedelta(params['start_date'] - start_year).total_seconds())
    nb_sec_per_day =  24. * 60. * 60.
    nb_sec_per_year = (365. * nb_sec_per_day)
    year_pattern = 2. * np.pi / nb_sec_per_year
    seasonal_pattern = 1.5 / 7. * np.cos(year_pattern * (t + start_min - 45 * nb_sec_per_day))  # min of the load is 15 of February so 45 days after beginning of year
    #seasonal_pattern = 1.5 / 7. * np.cos(year_pattern * (t - start_min - 30 * nb_sec_per_day)) #older version to be removed
    seasonal_pattern += 5.5 / 7.

    # Get weekly pattern
    weekly_pattern = compute_load_pattern(params, weekly_pattern, index, day_lag)
    std_temperature_noise = params['std_temperature_noise']
    residential_series = Pmax * weekly_pattern * (std_temperature_noise * temperature_signal + seasonal_pattern)

    return residential_series

def compute_load_pattern(params, weekly_pattern, index, day_lag):
    """
    Loads a typical hourly pattern, and interpolates it to generate
    a smooth solar generation pattern between 0 and 1

    Input:
        computation_params: (dict) Defines the mesh dimensions and
            precision. Also define the correlation scales
            interpolation_params: (dict) params of the interpolation

    Output:
        (np.array) A smooth solar pattern
    """
    # solar_pattern resolution : 1H, 8761

    if day_lag is None:
        nb_step_lag_for_starting_day = 0
    else:
        nb_step_lag_for_starting_day = 12 * 24 * day_lag
        
    # Keep only one week of pattern
    index_weekly_perweek = 12 * 24 * 7
    index %= int((weekly_pattern.shape[0] - nb_step_lag_for_starting_day) / index_weekly_perweek - 1)
    weekly_pattern = weekly_pattern[(nb_step_lag_for_starting_day + index * index_weekly_perweek):(nb_step_lag_for_starting_day + (index + 1) * index_weekly_perweek)]
    weekly_pattern /= np.mean(weekly_pattern)

    start_year = pd.to_datetime(str(params['start_date'].year) + '/01/01', format='%Y-%m-%d')
    T_bis = int(pd.Timedelta(params['end_date'] - start_year).total_seconds() // (60))

    Nt_inter_hr = int(T_bis // 5 + 1)
    N_repet = int((Nt_inter_hr - 1) // len(weekly_pattern) + 1)
    stacked_weekly_pattern = weekly_pattern
    for i in range(N_repet - 1):
        stacked_weekly_pattern = np.append(stacked_weekly_pattern, weekly_pattern)

    # The time is in minutes
    t_pattern = np.linspace(0, 60 * 7 * 24 * N_repet, 12 * 7 * 24 * N_repet, endpoint=False)
    f2 = interp1d(t_pattern, stacked_weekly_pattern, kind='cubic')

    Nt_inter = int(params['T'] // params['dt'] + 1)
    start_year = pd.to_datetime(str(params['start_date'].year) + '/01/01', format='%Y-%m-%d')
    start_min = int(pd.Timedelta(params['start_date'] - start_year).total_seconds() // 60)
    end_min = int(pd.Timedelta(params['end_date'] - start_year).total_seconds() // 60)
    t_inter = np.linspace(start_min, end_min, Nt_inter, endpoint=True)
    output = f2(t_inter)
    output = output * (output > 0)

    return output


def create_csv(prng, dict_, path, forecasted=False, reordering=True, noise=None,
               shift=False, write_results=True, index=False):
    df = pd.DataFrame.from_dict(dict_)
    df.set_index('datetime', inplace=True)
    df = df.sort_index(ascending=True)
    df = df.head(len(df)-1)  # Last value is lonely for another day
    if reordering:
        value = []
        for name in list(df):
            value.append(utils.natural_keys(name))
        new_ordering = [x for _, x in sorted(zip(value, list(df)))]
        df = df[new_ordering]
    if shift:
        df = df.shift(-1)
        df = df.fillna(0)

    df_reactive_power = 0.7 * df
    if noise is not None:
        df *= prng.lognormal(mean=0.0,sigma=noise, size=df.shape)
        #df *= np.random.lognormal(mean=0.0, sigma=noise, size=df.shape) #older version to be removed
        df_reactive_power *= prng.lognormal(mean=0.0, sigma=noise, size=df.shape)
        #df_reactive_power *= np.random.lognormal(mean=0.0, sigma=noise,
        #                                         size=df.shape) #older version to be removed

    if write_results:
        file_extension = '_forecasted' if forecasted else ''
        df.to_csv(
            os.path.join(path, f'load_p{file_extension}.csv.bz2'),
            index=index, sep=';', float_format=cst.FLOATING_POINT_PRECISION_FORMAT)
        df_reactive_power.to_csv(
            os.path.join(path, f'load_q{file_extension}.csv.bz2'),
            index=False, sep=';', float_format=cst.FLOATING_POINT_PRECISION_FORMAT)

    return df




# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

"""
Utility script to generate hydro guid curves (min, max constraint timeseries
for an economic dispatch problem) from an eco2mix excel file.
Beware that the downloaded .xls file has to be first opened in excel and saved
as a working xlsx file.
"""

import argparse
import datetime as dt

import pandas as pd


def q(percentile):
    """
    Generate a parametrized quantile function to apply over a numpy array
    Parameters.
    ----------
    percentile : float
            The percentile to use in the returned quantile function
    Returns
    -------
    The parametrized quantile function
    """
    def q_leveled(x):
        return x.quantile(percentile)

    # Rename the produced function to be able to use several functions in
    # pandas.DataFrame.groupby.agg method
    q_leveled.__name__ = 'q_{}'.format(str(percentile))

    return q_leveled


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description=('generate hydro guid curves (min, max constraint timeseries'
                     'for an economic dispatch problem) from an eco2mix excel file.'))
    parser.add_argument('filepath',
                        type=str,
                        help='Full path to the eco2mix excel file')
    parser.add_argument('--output', nargs=1, type=str,
                        default='output.csv',
                        help='File path for the output csv file (default: "output.csv")')

    args = parser.parse_args()

    if args.filepath.endswith('.xls'):
        print('Open first the .xls file in excel and save it as xlsx before' 
              'running the script')
        exit(0)

    # Read the eco2mix excel file and extract relevant data
    eco2mix = pd.read_excel(io=args.filepath)
    eco2mix = eco2mix[['Date', 'Heures', 'Hydraulique']]
    eco2mix = eco2mix.dropna()
    eco2mix.Heures = pd.to_datetime(eco2mix.Heures, format='%H:%M:%S')
    eco2mix['Datetime'] = eco2mix[['Date', 'Heures']].apply(
        lambda row: dt.datetime(row[0].year,
                                row[0].month,
                                row[0].day,
                                row[1].hour,
                                row[1].minute),
        axis=1
    )
    eco2mix = eco2mix[['Datetime', 'Hydraulique']]
    eco2mix = eco2mix.set_index('Datetime')

    # Normalize the data between 0 and 1
    eco2mix['Hydraulique'] = eco2mix['Hydraulique'] / eco2mix['Hydraulique'].max()

    # Recover the year of the eco2mix data
    year = eco2mix.index.year[0]

    # Recover monthly quantiles
    hydro_monthly = eco2mix.groupby(eco2mix.index.month).agg([q(0.05), q(0.95)])
    hydro_monthly.columns = hydro_monthly.columns.droplevel(0)

    output_range = pd.date_range(
        start=dt.datetime(year, 1, 1),
        end=dt.datetime(year, 12, 31, 23, 55),
        freq='5T')

    # Create the output file
    output = pd.DataFrame(index=output_range, columns=['Month'])
    output['Month'] = output.index.month
    output = output.merge(hydro_monthly, left_on='Month', right_index=True)
    output = output.drop('Month', axis=1)
    output = output.merge(eco2mix, left_index=True, right_index=True, how='left')
    output = output.fillna(method='ffill')

    output.to_csv(args.output[0], sep=',', decimal='.')



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import datetime as dt
import math
import os

import pandas as pd
import pathlib

from .generation import generation_utils as gu
from chronix2grid import constants as cst


def write_start_dates_for_chunks(output_path, scenario_name, n_weeks, by_n_weeks,
                                 n_scenarios, start_date, time_step):
    days_to_day = by_n_weeks * 7
    n_chunks = compute_n_chunks(n_weeks, by_n_weeks)
    start_date_time = pd.to_datetime(start_date, format='%Y-%m-%d')
    file_name = 'start_datetime.info'

    if time_step % 60 == 0:
        hours = time_step // 60
        time_step = pd.to_datetime(f'0'+str(hours)+':00', format='%H:%M')
    else:
        time_step = pd.to_datetime(f'00:{str(time_step)}', format='%H:%M')
    file_name_ts = cst.TIME_STEP_FILE_NAME
    
    scen_name_generator = gu.folder_name_pattern(scenario_name, n_scenarios)
    
    chunk_folder_name_generator = gu.folder_name_pattern('chunk', n_chunks)
    for j in range(n_scenarios):
        if (n_scenarios>1):#otherwise keep scenario_name as defined
            scenario_name = scen_name_generator(j)
        write_start_datetime_info(
            os.path.join(output_path, scenario_name),
            file_name,
            start_date_time)
        write_timestep_info(os.path.join(output_path, scenario_name), file_name_ts, time_step)
        for i in range(n_chunks):
            chunk_folder_name = chunk_folder_name_generator(i)
            output_directory = os.path.join(
                output_path, scenario_name, chunk_folder_name)
            write_start_datetime_info(output_directory, file_name, start_date_time)
            write_timestep_info(output_directory, file_name_ts, time_step)
            start_date_time += dt.timedelta(days=days_to_day)


def write_start_datetime_info(directory_path, file_name, start_date_time):
    with open(os.path.join(directory_path, file_name), 'w') as f:
        f.write(start_date_time.strftime('%Y-%m-%d %H:%M'))
        
def write_timestep_info(directory_path, file_name, time_interval):
    with open(os.path.join(directory_path, file_name), 'w') as f:
        f.write(time_interval.strftime('%H:%M'))


def compute_n_chunks(n_weeks, by_n_weeks):
    n_chunks = 0
    if n_weeks > by_n_weeks:
        n_chunks = math.ceil(n_weeks / by_n_weeks)
    return n_chunks


def output_processor_to_chunks(output_path, scenario_name, by_n_weeks, n_scenarios, n_weeks):
    if n_weeks > by_n_weeks:
        chunk_size = by_n_weeks * 7 * 24 * 12  # 5 min time step
    
        scen_name_generator = gu.folder_name_pattern(scenario_name, n_scenarios)
        for i in range(n_scenarios):
            if (n_scenarios>1):#otherwise keep scenario_name as defined
                scenario_name = scen_name_generator(i)
            csv_files_to_process = os.listdir(os.path.join(output_path, scenario_name))
            csv_files_to_process = [
                os.path.join(output_path, scenario_name, csv_file) for csv_file in csv_files_to_process
                if os.path.isfile(os.path.join(output_path, scenario_name, csv_file))
            ]
            generate_chunks(csv_files_to_process, chunk_size)


def generate_chunks(csv_files_to_process, chunk_size, sep=','):
    for csv_file in csv_files_to_process:
        cut_df = cut_csv_file_into_chunks(csv_file, chunk_size, sep=sep)
        save_chunks(cut_df, csv_file, index=False)


def cut_csv_file_into_chunks(csv_file_path, chunk_size, **kwargs):
    df = pd.read_csv(csv_file_path, **kwargs)
    return dataframe_cutter(df, chunk_size)


def save_chunks(chunks, original_file_path, **kwargs):
    parent_dir = pathlib.Path(original_file_path).parent.absolute()
    original_file_name = pathlib.Path(original_file_path).name
    chunk_folder_name_generator = gu.folder_name_pattern('chunk', len(chunks))
    for i, chunk in enumerate(chunks):
        chunk_folder_name = chunk_folder_name_generator(i)
        os.makedirs(os.path.join(parent_dir, chunk_folder_name), exist_ok=True)
        chunk.to_csv(
            os.path.join(parent_dir, chunk_folder_name, original_file_name),
            **kwargs
        )


def dataframe_cutter(df, chunk_size):
    """
    Cut a csv file into chunks
    Parameters
    ----------
    df pandas.DataFrame
        The DataFrame to cut
    chunk_size int
        The size of each chunk

    Returns
    -------
        The list of cut dataframes
    """
    cut_df = []
    if chunk_size < len(df):
        n_chunks = len(df) // chunk_size
        for i in range(n_chunks):
            cut_df.append(df.iloc[i*chunk_size:((i+1)*chunk_size)])
        if len(df) > chunk_size * n_chunks:
            cut_df.append(df.iloc[chunk_size*n_chunks:])
    return cut_df



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import argparse
import numpy as np
import pandas as pd
import seaborn as sns
from itertools import product
from pathlib import Path
from matplotlib import pyplot as plt
#from entropy import spectral_entropy

class HydroKPI:

    def __init__(self, kpi_databanks):
        self.kpi_databanks = kpi_databanks

    def __hydro_in_prices(self,
                          norm_mw,
                          upper_quantile,
                          lower_quantile,
                          above_norm_cap,
                          below_norm_cap):

        '''
        '''

        # Get number of gen units
        n_units = norm_mw.shape[1]

        # Get the price at upper/lower quantile
        eu_upper_quantile = self.prices.quantile(upper_quantile)
        eu_lower_quantile = self.prices.quantile(lower_quantile)

        # Test if units are above/below normalized capacity
        is_gens_above_cap =  norm_mw.ge(eu_upper_quantile, axis=1)
        is_gens_below_cap =  norm_mw.le(eu_lower_quantile, axis=1)

        # Test if prices are greater/lower than defined quantiles
        is_price_above_q = self.prices > eu_upper_quantile
        is_price_below_q = self.prices < eu_lower_quantile

        # Stacking price bool condition for all hydro units along
        # the columns and rename them as units names
        is_price_above_q_gens = is_price_above_q[['price'] * n_units]
        is_price_above_q_gens.columns = norm_mw.columns

        is_price_below_q_gens = is_price_below_q[['price'] * n_units]
        is_price_below_q_gens.columns = norm_mw.columns

        # Match occurence of price high and full gen disaptch
        high_price_kpi = 100 * (is_price_above_q_gens & is_gens_above_cap).sum(axis=0) \
                         / is_price_above_q_gens.sum(axis=0)

        # Match occurence when price is lower and almost no dispatch
        low_price_kpi = 100 * (is_price_below_q_gens & is_gens_below_cap).sum(axis=0) \
                        / is_price_below_q_gens.sum(axis=0)

        return high_price_kpi.round(self.precision), low_price_kpi.round(self.precision)

    def __hydro_seasonal(self, hydro_mw):

        '''
        '''

        # We aggregate hydro per month if a user want to deliver less MW in some
        # months rather than others.
        #
        # E.g. Typical configuration:
        # 6 month full capacity and
        # other months with 0 MW)
        hydro_mw_month = hydro_mw.copy()
        hydro_mw_month['month'] = self.months
        mw_per_month = hydro_mw_month.groupby('month').sum(axis=0).round(self.precision)

        return mw_per_month

    def hydro_kpi(self,
                  upper_quantile = 0.95,
                  lower_quantile = 0.05,
                  above_norm_cap = 0.9,
                  below_norm_cap = 0.1):

        '''
        Get 4 different Hydro KPI's based on the assumption the total costs
        of the system follow same curve as the consumption.

        Parameters:
        ----------

        upper_quantile (float): Quantile that define high prices.
                                (Prices are considered high whether
                                price(t) is greater than upper_quantile)
        lower_quantile (float): Quantile that define lower prices
                                (Prices are considered low whether
                                price(t) is less than lower_quantile)

        above_cap (float): Per unit (<1) criteria to establish high hydro dispatch
        below_cap (float): Per unit (<1) criteria to establish low hydro dispatch

        Returns:
        --------

        highPrice_kpi (dataframe): Percentage of time a generator is keeping
                                   operating above its predefined capacity

        lowPrice_kpi (dataframe): Percentage of time a generator is keeping
                                   operating below its predefined capacity

        mw_per_month (dataframe): Aggregated sum per month. Used to design
                                  seasonal pattern in hydro plants.
        '''

        # Get Hydro names
        hydro_filter = self.prod_charac.type.isin(['hydro'])
        hydro_names = self.prod_charac.name.loc[hydro_filter].values

        # Normalize MW according to the max value for
        # the reference data and synthetic one
        hydro_ref = self.ref_dispatch[hydro_names]
        hydro_syn = self.syn_dispatch[hydro_names]

        max_mw_ref = hydro_ref.max(axis=0)
        max_mw_syn = hydro_syn.max(axis=0)

        norm_mw_ref = hydro_ref / max_mw_ref
        norm_mw_syn = hydro_syn / max_mw_syn

        # Stats for reference data
        stat_ref_high_price, stat_ref_low_price = self.__hydro_in_prices(norm_mw_ref,
                                                                         upper_quantile,
                                                                         lower_quantile,
                                                                         above_norm_cap,
                                                                         below_norm_cap)

        # Stats for synthetic data
        stat_syn_high_price, stat_syn_low_price = self.__hydro_in_prices(norm_mw_syn,
                                                                         upper_quantile,
                                                                         lower_quantile,
                                                                         above_norm_cap,
                                                                         below_norm_cap)

        # Write results
        # -- + -- + --
        self.output['Hydro'] = {'high_price_for_ref': stat_ref_high_price.to_dict(),
                                'low_price_for_ref': stat_ref_low_price.to_dict(),
                                'high_price_for_syn': stat_syn_high_price.to_dict(),
                                'low_price_for_syn': stat_syn_low_price.to_dict()
                                }

        # Seasonal for reference data
        ref_agg_mw_per_month = self.__hydro_seasonal(hydro_ref)

        # Seasonal for synthetic data
        syn_agg_mw_per_month = self.__hydro_seasonal(hydro_syn)

        # Write results
        # -- + -- + --
        self.output['Hydro'] = {'seasonal_month_for_ref': ref_agg_mw_per_month.to_dict(),
                                'seasonal_month_for_syn': syn_agg_mw_per_month.to_dict()
                                }

        return stat_ref_high_price, stat_ref_low_price, ref_agg_mw_per_month, \
               stat_syn_high_price, stat_syn_low_price, syn_agg_mw_per_month
